from traits.api import *
from traitsui.api import *#View, Item, ButtonEditor, Group, HSplit,
from traitsui.menu import *

import pylab as plb
import skrf as rf
import os
from plotTool import PlotTool

class Network2Spreadsheet(HasTraits):
    input_dir = Directory('')
    output_dir = Directory('')

    output_filename = String('output.xls')
    ntwk_dict = Property(depends_on = 'input_dir')
    available_ntwk_list = Property(depends_on = 'input_dir')
    active_ntwk_name = Enum(values='available_ntwk_list')
    active_ntwk = Property(depends_on = 'active_ntwk_name')

    export_network = Button('Export Network')
    export_all_networks = Button('Export All Networks')
    form = Enum('dB/deg',['dB/deg','mag/deg','real/imag'])

    @cached_property
    def _get_ntwk_dict(self):
        try:
            return rf.ran(self.input_dir)
        except(OSError):
            return {}

    def _get_available_ntwk_list(self):
        return  self.ntwk_dict.keys()

    def _get_active_ntwk(self):
        try:
            return self.ntwk_dict[self.active_ntwk_name]
        except:
            pass

    def _get_ports(self):
        try:
            return ['All']+range(self.active_ntwk.nports)
        except:
            return []

    def _export_network_fired(self):
        form_dict = {
            'dB/deg':'db',
            'mag/deg':'ma',
            'real/imag':'ri',
            }
        form = form_dict[self.form]
        file_name = os.path.join(self.output_dir, self.output_filename)

        rf.io.general.network_2_spreadsheet(
            self.active_ntwk,
            file_name = file_name,
            form = form )
        print(('{} --> {}'.format(self.active_ntwk_name,file_name)))

    def _export_all_networks_fired(self):
        form_dict = {
            'dB/deg':'db',
            'mag/deg':'ma',
            'real/imag':'ri',
            }
        form = form_dict[self.form]
        file_name = os.path.join(self.output_dir, self.output_filename)
        ns = rf.NS(self.ntwk_dict)
        rf.io.general.networkset_2_spreadsheet(
            ns,
            file_name=file_name,
            form = form
            )
        for k in [ '{} --> {}'.format(k.name,file_name) for k in ns.ntwk_set]:
            print(k)


    view = View(
        Group(
            Item('input_dir', label='Input Dir'),
            Item('output_dir', label='Output Dir'),
            Item('output_filename', label='Output Filename'),
            HSplit(
                Item('active_ntwk_name', show_label=False),
                Item('export_network',show_label=False),
                Item('export_all_networks',show_label=False),
                Item('form')

                ),
            ),
        resizable=True,
        width = 800, height = 300,
        title = 'Skrf Spreadsheet Exporter'
        )


if __name__  == '__main__':
    a = Network2Spreadsheet()
    a.configure_traits()



"""
.. module:: skrf.time
========================================
time (:mod:`skrf.time`)
========================================

time domain functions 

Methods
------------

.. autosummary::
    :toctree: generated/

    time_gate
    detect_span 
    find_n_peaks
    indexes
    

    
"""

from .util import  find_nearest_index


from scipy.ndimage.filters import convolve1d
from scipy import  signal
import numpy as npy
import numpy as np # so i dont have to change indexes (from peakutils)
from numpy import fft

def indexes(y, thres=0.3, min_dist=1):
    """Peak detection routine.

    Finds the numeric index of the peaks in *y* by taking its first order difference. By using
    *thres* and *min_dist* parameters, it is possible to reduce the number of
    detected peaks. *y* must be signed.

    Parameters
    ----------
    y : ndarray (signed)
        1D amplitude data to search for peaks.
    thres : float between [0., 1.]
        Normalized threshold. Only the peaks with amplitude higher than the
        threshold will be detected.
    min_dist : int
        Minimum distance between each detected peak. The peak with the highest
        amplitude is preferred to satisfy this constraint.

    Returns
    -------
    ndarray
        Array containing the numeric indexes of the peaks that were detected
    
    Notes
    --------
    This function was taken from peakutils-1.1.0 
    http://pythonhosted.org/PeakUtils/index.html
    
    """
    #This function  was taken from peakutils, and is covered 
    # by the MIT license, inlcuded below: 
    
    #The MIT License (MIT)

    #Copyright (c) 2014 Lucas Hermann Negri

    #Permission is hereby granted, free of charge, to any person obtaining a copy
    #of this software and associated documentation files (the "Software"), to deal
    #in the Software without restriction, including without limitation the rights
    #to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    #copies of the Software, and to permit persons to whom the Software is
    #furnished to do so, subject to the following conditions:

    #The above copyright notice and this permission notice shall be included in
    #all copies or substantial portions of the Software.

    #THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    #IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    #FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    #AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    #LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    #OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
    #THE SOFTWARE.
    
    if isinstance(y, np.ndarray) and np.issubdtype(y.dtype, np.unsignedinteger):
        raise ValueError("y must be signed")

    thres = thres * (np.max(y) - np.min(y)) + np.min(y)
    min_dist = int(min_dist)

    # compute first order difference
    dy = np.diff(y)

    # propagate left and right values successively to fill all plateau pixels (0-value)
    zeros,=np.where(dy == 0)
    
    # check if the singal is totally flat
    if len(zeros) == len(y) - 1:
        return np.array([])
    
    while len(zeros):
        # add pixels 2 by 2 to propagate left and right value onto the zero-value pixel
        zerosr = np.hstack([dy[1:], 0.])
        zerosl = np.hstack([0., dy[:-1]])

        # replace 0 with right value if non zero
        dy[zeros]=zerosr[zeros]
        zeros,=np.where(dy == 0)

        # replace 0 with left value if non zero
        dy[zeros]=zerosl[zeros]
        zeros,=np.where(dy == 0)

    # find the peaks by using the first order difference
    peaks = np.where((np.hstack([dy, 0.]) < 0.)
                     & (np.hstack([0., dy]) > 0.)
                     & (y > thres))[0]

    # handle multiple peaks, respecting the minimum distance
    if peaks.size > 1 and min_dist > 1:
        highest = peaks[np.argsort(y[peaks])][::-1]
        rem = np.ones(y.size, dtype=bool)
        rem[peaks] = False

        for peak in highest:
            if not rem[peak]:
                sl = slice(max(0, peak - min_dist), peak + min_dist + 1)
                rem[sl] = True
                rem[peak] = False

        peaks = np.arange(y.size)[~rem]

    return peaks

def find_n_peaks(x,n, thres=.9, **kw):
    '''
    Find a given number of peaks in a signal
    '''
    
    for dummy  in range(10):
        
        idx = indexes(x, **kw)
        if len(idx) < n:
            thres*=.5
            
        else:
            peak_vals = sorted(x[idx], reverse=True)[:n]
            peak_idxs =[x.tolist().index(k) for k in peak_vals]

            return peak_idxs
    raise ValueError('Couldnt find %i peaks'%n)
    
def detect_span(ntwk):
    '''
    detect the correct time-span between two largest peaks
    '''
    x = ntwk.s_time_db.flatten()
    p1,p2 = find_n_peaks(x,n=2)
    #distance to nearest neighbor peak
    span = abs(ntwk.frequency.t_ns[p1]-ntwk.frequency.t_ns[p2])
    return span 
    
def time_gate(ntwk, start=None, stop=None, center=None, span=None,
              mode='bandpass', window=('kaiser', 6),media=None, 
              boundary='reflect',return_all=False):
    '''
    Time-gate one-port s-parameters.
    
    The gate can be defined with start/stop times, or by 
    center/span. all times are in units of nanoseconds. common 
    windows are:
     * ('kaiser', 6)
     * 6 # integers are interpreted as kaiser beta-values
     * 'hamming'
     * 'boxcar'  # a staightup rect
     
    If no parameters are passed this will try to auto-gate the largest
    peak. 

    Parameters
    ------------
    start : number, or None
        start of time gate, (ns). 
    stop : number, or None
        stop of time gate (ns). 
    center : number, or None
        center of time gate, (ns). If None, and span is given, 
        the gate will be centered on the peak.
    span : number, or None
        span of time gate, (ns).  If None span will be half of the 
        distance to the second tallest peak
    mode: ['bandpass','bandstop']
        mode of gate 
    boundary:  {'reflect', 'constant', 'nearest', 'mirror', 'wrap'},
        passed to `scipy.ndimage.filters.convolve1d`
    window : string, float, or tuple 
        passed to `window` arg of `scipy.signal.get_window`
    
    Notes
    ------
    You cant gate things that are 'behind' strong reflections. This 
    is due to the multiple reflections that occur. 
    
    If `center!=0`, then the ntwk's time response is shifted 
    to t=0, gated, then shifted back to where it was. This is 
    done in frequency domain using `ntwk.delay()`. If the media being
    gated is dispersive (ie waveguide), then the gate `span` will be 
    span at t=0, which is different.
    
    If you need to time-gate an N-port network, then you should 
    gate each s-parameter independently. 
    
    Returns
    --------
    ntwk : Network
        copy of ntwk with time-gated s-parameters

    .. warning::
        Depending on sharpness of the gate, the  band edges may be 
        inaccurate, due to properties of FFT. We do not re-normalize
        anything.


    '''
    if ntwk.nports >1:
        raise ValueError('Time-gating only works on one-ports. try taking ntwk.s11 or ntwk.s21 first')

    if start is not None and stop is not None:
        start *= 1e-9
        stop *= 1e-9
        span = abs(stop-start)
        center = (stop+start)/2.
    
    else:
        if center is None:    
            # they didnt provide center, so find the peak
            n = ntwk.s_time_mag.argmax()
            center = ntwk.frequency.t_ns[n]
            
        if span is None:
            span = detect_span(ntwk)
            
        center *= 1e-9
        span *= 1e-9
        start = center - span / 2.
        stop = center + span / 2.

    
    # find start/stop gate indecies
    t = ntwk.frequency.t
    start_idx = find_nearest_index(t, start)
    stop_idx = find_nearest_index(t, stop)

    # create window
    window_width = abs(stop_idx - start_idx)
    window = signal.get_window(window, window_width)

    # create the gate by padding the window with zeros
    gate = npy.r_[npy.zeros(start_idx),
                           window,
                           npy.zeros(len(t) - stop_idx)]

    #IFFT the gate, so we have it's frequency response, aka kernel
    kernel=fft.fftshift(fft.ifft(fft.ifftshift(gate, axes=0), axis=0))
    kernel =abs(kernel).flatten() # take mag and flatten
    kernel=kernel/sum(kernel) # normalize kernel
    
    out = ntwk.copy()
    
    # conditionally delay ntwk, to center at t=0, this is 
    # equivalent to gating at center.  (this is probably very inefficient)
    if center!=0:
        out = out.delay(-center*1e9, 'ns',port=0,media=media)
    
    # waste of code to handle convolve1d suck
    re = out.s_re[:,0,0]
    im = out.s_im[:,0,0]
    s = convolve1d(re,kernel, mode=boundary)+\
     1j*convolve1d(im,kernel, mode=boundary)
    out.s[:,0,0] = s
    # conditionally  un-delay ntwk
    if center!=0:
        out = out.delay(center*1e9, 'ns',port=0,media=media)

    if mode == 'bandstop':
        out = ntwk-out
    elif mode=='bandpass':
        pass
    else:
        raise ValueError('mode should be \'bandpass\' or \'bandstop\'')

    if return_all:
        # compute the gate ntwk and add delay
        gate_ntwk = out.s11.copy()
        gate_ntwk.s = kernel
        gate_ntwk= gate_ntwk.delay(center*1e9, 'ns', media=media)
        
        return {'gated_ntwk':out,
                'gate':gate_ntwk}
    else:
        return out
    






"""
This is a model module.  It will not function correctly to pull data, but needs
to be subclassed.
"""
import copy
import warnings
from typing import Iterable

import numpy as np
import pyvisa

from ...network import Network
from ...frequency import Frequency


class VNA(object):
    """
    class defining a base analyzer for using with scikit-rf

    This class defines the base functionality expected for all Network
    Analyzers. To keep this manageable, it defines only the most commonly used
    applications, which currently are:
    * grabbing data (SNP networks, and traces)
    * grabbing a catalogue of available traces
    * instantiating and grabbing switch terms for calibration
    * setting the frequency parameters (start, stop, number of points, type)

    Attributes
    ----------
    idn
    nports

    Methods
    -------
    get_oneport(ports=(1, 2), **kwargs)
        get an return a 1-port Network object
    get_twoport(ports=1, **kwargs)
        get an return a 2-port Network object
    enter/exit - for using python's with statement
    >>> with Analyzer("GPIB::16::ISNTR") as nwa:
    >>>     ntwk = nwa.measure_twoport_ntwk()

    Methods that must be subclassed
    -------------------------------
    get_traces
    get_list_of_traces
    get_snp_network
    get_switch_terms
    set_frequency_sweep
    """

    def __init__(self, address, **kwargs):
        """
        Initialize a network analyzer object

        Parameters
        ----------
        address : str
            a visa resource string, or an ip address
        kwargs : dict
            visa_library (str), timemout in milliseconds (int), card_number
            (int), interface (str)

        Notes
        -----
        General initialization of a skrf vna object.  Defines the class methods
        that all subclasses should provide, and therefore defines the basic
        functionality, currently focused on grabbing data more than on
        controlling the state of the vna

        Keyword Arguments
        -----------------
        visa_library : str
            allows pyvisa to use different visa_library backends, including the
            python-based pyvisa-py.  backend which can handle SOCKET and Serial
            (though not GPIB) connections.  It should be possible to use this
            library without NI-VISA libraries installed if the analyzer is so
            configured.
        timeout : int
            milliseconds
        interface : str
            one of "SOCKET", "GPIB"
        card_number : int
            for GPIB, default is usually 0
        """

        rm = kwargs.get("resource_manager", None)
        if not rm:
            rm = pyvisa.ResourceManager(visa_library=kwargs.get("visa_library", ""))

        interface = str(kwargs.get("interface", None)).upper()  # GPIB, SOCKET
        if interface == "GPIB":
            board = str(kwargs.get("card_number", "")).upper()
            resource_string = "GPIB{:}::{:}::INSTR".format(board, address)
        elif interface == "SOCKET":
            port = str(kwargs.get("port", 5025))
            resource_string = "TCPIP0::{:}::{:}::SOCKET".format(address, port)
        else:
            resource_string = address
        self.resource = rm.open_resource(resource_string)  # type: pyvisa.resources.messagebased.MessageBasedResource
        self.resource.timeout = kwargs.get("timeout", 3000)

        self.resource.read_termination = "\n"  # most queries are terminated with a newline
        self.resource.write_termination = "\n"
        if "instr" in resource_string.lower():
            self.resource.control_ren(2)

        # convenience pyvisa functions
        self.write = self.resource.write
        self.read = self.resource.read
        self.query = self.resource.query
        self.query_values = self.resource.query_values

    def __enter__(self):
        """
        context manager entry point

        Returns
        -------
        VNA
            the Analyzer Driver Object
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        context manager exit point

        Parameters
        ----------
        exc_type : type
        exc_val : type
        exc_tb : traceback
        """
        self.resource.close()

    def close(self):
        self.__exit__(None, None, None)

    @property
    def idn(self):
        return self.query("*IDN?")

    def reset(self):
        self.write("*RST")

    def wait_until_finished(self):
        self.query("*OPC?")

    def get_list_of_traces(self, **kwargs):
        """
        a catalogue of the available data traces

        Parameters
        ----------
        kwargs : dict
            to allow for optional parameters in the sub-class

        Returns
        -------
        list
            catalogue of the available traces with description and all
            information necessary to index the trace and grab it from the
            analyzer (i.e. channel, name, parameter etc.

        Notes
        -----
        the purpose of this function is to query the analyzer for the available
        traces, and then it then returns a list where each list-item represents
        one available trace.  How this is achieved is completely up to the user
        with the only requirement that the items from this list must be passed
        to the self.get_traces function, which will return one network item for
        each item in the list that is passed.

        Typically the user will get this list of all available traces, and then
        by some functionality in the widgets (or whatever) will down-select the
        list and then return that down-selected list to the get_traces function
        to retrieve the desired networks.

        Each list item then must be a python object (str, list, dict, etc.) with
        all necessary information to retrieve the trace as an Network object.
        For example, each item could be a python dict with the following keys:
        * "name": the name of the measurement e.g. "CH1_S11_1"
        * "channel": the channel number the measurement is on
        * "measurement": the measurement number (MNUM in SCPI)
        * "parameter": the parameter of the measurement, e.g. "S11", "S22" or
          "a1b1,2"
        * "label": the text the item will use to identify itself to the user e.g
          "Measurement 1 on Channel 1"
        """
        raise NotImplementedError("must implement with subclass")

    def get_traces(self, traces, **kwargs):
        """
        retrieve traces as 1-port networks from a list returned by
        get_list_of_traces

        Parameters
        ----------
        traces : list
            list of type that is exported by self.get_list_of_traces
        kwargs : dict
            to provide for optional parameters

        Returns
        -------
        list
            a list of 1-port networks representing the desired traces

        Notes
        -----
        There is no current way to distinguish between traces and 1-port
        networks within skrf
        """
        raise NotImplementedError("must implement with subclass")

    def get_snp_network(self, ports, **kwargs):
        """
        return n-port network as an Network object

        Parameters
        ----------
        ports : Iterable
            a iterable of integers designating the ports to query
        kwargs : dict
            optional arguments for subclassing

        Returns
        -------
        Network

        general function to take in a list of ports and return the full snp
        network as a Network object
        """
        raise NotImplementedError("must implement with subclass")

    def get_twoport(self, ports=(1, 2), **kwargs):
        """
        convenience wrapper for get_snp_network enforcing 2-ports only

        Parameters
        ----------
        ports : Iterable
            a 2-length iterable of integers specifying the ports
        kwargs : dict
            option parameters, i.e. channel, sweep, etc.

        Returns
        -------
        Network
        """
        if len(ports) != 2:
            raise ValueError("Must provide a 2-length list of integers for the ports")
        return self.get_snp_network(ports, **kwargs)

    def get_oneport(self, port=1, **kwargs):
        """
        convenience wrapper for get_snp_network enforcing 1-port only

        Parameters
        ----------
        ports : int
            integer specifying the port number
        kwargs : dict
            option parameters, i.e. channel, sweep, etc.

        Returns
        -------
        Network
        """
        if type(port) in (list, tuple):
            if len(port) > 1:
                raise ValueError("specify the port as an integer")
        else:
            if type(port) is int:
                port = (port,)
            else:
                raise ValueError("specify the port as an integer")

        return self.get_snp_network(port, **kwargs)

    def get_switch_terms(self, ports=(1, 2), **kwargs):
        """
        create new traces for the switch terms and return as a 2-length list of
        forward and reverse terms

        Parameters
        ----------
        ports : Iterable
            a 2-length iterable of integers specifying the ports

        kwargs : dict
            optional parameters

        Returns
        -------
        list
            a 2-length list of 1-port networks [forward_switch_terms,
            reverse_switch_terms]
        """
        raise NotImplementedError("must implement with subclass")

    def set_frequency_sweep(self, start_freq, stop_freq, num_points, **kwargs):
        """
        Set the frequency sweep parameters on the specified or active channel

        Parameters
        ----------
        start_freq : float
            the starting frequency in f_units (Hz, GHz, etc.)
        stop_freq : float
            the ending frequency in f_units (Hz, GHz, etc.)
        num_points : int
            the number of points in the frequency sweep
        kwargs : dict
            channel (int), f_units (str), sweep_type (str -> lin, log), other
            optional parameters
        """
        raise NotImplementedError("must implement with subclass")

    @staticmethod
    def to_hz(freq, f_unit):
        """
        A simple convenience function to create frequency in Hz if it is in a
        different unit

        Parameters
        ----------
        freq : float or np.ndarray
            a float or numpy.ndarray of floats of the frequency in f_units
        f_unit : str
            the units of frequency (Hz, kHz, MHz, GHz, THz)

        Returns
        -------
        float or np.ndarray
            the converted frequency sweep in Hz
        """
        return freq * Frequency.multiplier_dict[f_unit.lower()]





'''

.. currentmodule:: skrf.constants
========================================
constants (:mod:`skrf.constants`)
========================================

This module contains constants, numerical approximations, and unit conversions


.. data:: INF

    A very very large value (1e99)

.. data:: ONE

    1 + epsilon where epsilon is small. Used to avoid numerical error.

.. data:: ZERO

    0 + epsilon where epsilon is small. Used to avoid numerical error.

.. data:: K_BOLTZMANN

    Boltzmann constant (1.38064852e-23)

.. data:: S_DEFINITIONS

    S-parameter definition labels: 
        - 'power' for power-waves definition, 
        - 'pseudo' for pseudo-waves definition. 
        - 'traveling' corresponds to the initial implementation. 

.. data:: S_DEF_DEFAULT

    Default S-parameter definition: 'power', for power-wave definition.

.. autosummary::
   :toctree: generated/

   to_meters

'''

from scipy.constants import c, micron, mil, inch, centi, milli, nano, micro,pi



# used as substitutes to handle mathematical singularities.
INF = 1e99
ONE = 1.0 + 1/1e14
ZERO = 1e-6

K_BOLTZMANN = 1.38064852e-23
T0 = 290.

# S-parameter definition labels and default definition
S_DEFINITIONS = ['power', 'pseudo', 'traveling']
S_DEF_DEFAULT = 'power'

global distance_dict
distance_dict = {'m':1.,
                 'cm':1e-2,
                 'mm':1e-3,
                 'um':1e-6,
                 'in':inch,
                 'mil': mil,
                 's':c,
                 'us':1e-6*c,
                 'ns':1e-9*c,
                 'ps':1e-12*c,
                 }

def to_meters( d, unit='m',v_g=c):
    '''
    Translate various  units of distance into meters

    

    Parameters
    ------------
    d : number or array-like
        the value
    unit : str
        the unit to that x is in:
        ['m','cm','um','in','mil','s','us','ns','ps']
    v_g : 

    '''
    
    distance_dict = {'m':1.,
                 'cm':1e-2,
                 'mm':1e-3,
                 'um':1e-6,
                 'in':inch,
                 'mil': mil,
                 's':v_g,
                 'us':1e-6*v_g,
                 'ns':1e-9*v_g,
                 'ps':1e-12*v_g,
                 }
                 
                 
    unit = unit.lower()
    try:
        return distance_dict[unit]*d
    except(KeyError):
        raise(ValueError('Incorrect unit'))






'''
.. currentmodule:: skrf.media.device

========================================
device (:mod:`skrf.media.device`)
========================================

Device is a generic n-port microwave device class to create common Network.

Device Class
---------------
.. autosummary::
    :toctree: generated/

    Device

Example Devices
---------------
.. autosummary::
    :toctree: generated/

    MatchedSymmetricCoupler
    Hybrid
    QuadratureHybrid
    Hybrid180
    DualCoupler

'''


import numpy as npy
from abc import ABCMeta, abstractmethod, abstractproperty
from .. import mathFunctions as mf
from ..network  import connect
from numpy import sqrt,exp


class Device(object):
    '''
    A n-port microwave device
    
    Parameters 
    -----------
    media : skrf.media.Media
    '''
    __metaclass__ = ABCMeta
    
    def __init__(self, media):
        self.media = media 
        
    @abstractproperty
    def ntwk(self):
        '''
        the network representation of a given device
        '''
        return None
        
        
class MatchedSymmetricCoupler(Device):
    '''
    A Matched Symmetric Coupler
    
    The resultant ntwk port assignment is as follows:
        * 0 - insertion
        * 1 - transmit
        * 2 - coupled 
        * 3 - isolated
    '''
    def __init__(self, media, c=None, t=None, t_phase=0, phase_diff=0, 
                 nports=4, *args, **kw):
        Device.__init__(self, media=media, *args, **kw)
        
        if c is None and t is None:
            raise ValueError('Must pass either `c`  or `t`')
        
        if nports not in [3,4]:
            raise ValueError('nports must be 3 or 4')
        
        self.phase_diff = phase_diff
        self.t_phase = t_phase
        self.i = 0
        self.nports=nports
        
        
        if c is not None:
            self.c = c
        
        if t is not None:
            self.t = t
        
  
    @classmethod
    def from_dbdeg(cls, media, db, deg=0, *args,**kw):
        '''
        Create a coupler in terms of couping(dB) and phase offset(deg)
        
        Parameters 
        -----------
        media : skrf.media.Media Object
        
        db : number or array-like
            the magnitude of the coupling value (in dB), sign dont matter
        deg : number or array-like
            phase offset between the transmit and coupled arms. defined
            as : coupled arm = transmit arm +phase offset
        \*args,\*\*kw: passed to self.__init__()    
        
        '''
        c = mf.db_2_mag(-1*abs(db))
        return cls(media=media, c=c, phase_diff=deg, *args,**kw)
    
    @property
    def c(self):
        return self._c
    
    @c.setter
    def c(self,c):
        self._c = c
        t_rad = mf.degree_2_radian(self.t_phase)
        c_rad = t_rad + mf.degree_2_radian(self.phase_diff)
        
        self._t = npy.sqrt(1- npy.abs(c)**2)*exp(1j*t_rad)
        self._c = c*exp(1j*c_rad)
    
    @property
    def t(self):
        return self._t
    
    @t.setter
    def t(self,t):
        t_rad = mf.degree_2_radian(self.t_phase)
        c_rad = t_rad + mf.degree_2_radian(self.phase_diff)
        self._t= t*exp(1j*t_rad)
        self._c = npy.sqrt(1- npy.abs(t)**2)*exp(1j*c_rad)
    
    
    @property
    def ntwk(self):
        a = self.media.match(nports=4)
        a.s[:,0,1] = a.s[:,1,0] = a.s[:,3,2] = a.s[:,2,3] = self.t
        a.s[:,0,2] = a.s[:,2,0] = a.s[:,3,1] = a.s[:,1,3] = self.c
        a.s[:,0,3] = a.s[:,3,0] = a.s[:,1,2] = a.s[:,2,1] = self.i
        if self.nports ==3:
            match = self.media.match()
            a = connect(a,3,match,0)
        return a

    
class Hybrid(MatchedSymmetricCoupler):
    '''
    A 3dB Coupler of given phase difference
    '''
    def __init__(self, media, t_phase=180,phase_diff=0, *args, **kw):
        c = 1/sqrt(2)
        MatchedSymmetricCoupler.__init__(self,media=media, c=c, 
                                         phase_diff=phase_diff, 
                                         t_phase=t_phase,*args,**kw)


        
class QuadratureHybrid(MatchedSymmetricCoupler):
    '''
    A 3dB Coupler with 90deg phase diff between transmit and coupled arms
    '''
    def __init__(self, media,t_phase=0, *args, **kw):
        c = 1/sqrt(2)
        MatchedSymmetricCoupler.__init__(self,media=media,c=c,t_phase=t_phase,
                                         phase_diff=-90, *args, **kw)

    
class Hybrid180(Device):
    '''
    180degree hybrid
    
    This device can be used to combine two signals  in and out of phase,
    or as a divider, with outputs in or out of phase.
    
    The resultant ntwk port assignment is as follows:
        * 0 - sum (A+B)
        * 1 - input A
        * 2 - input B 
        * 3 - delta (A-B)
        
        
    http://www.microwaves101.com/encyclopedias/hybrid-couplers
    '''
    def __init__(self, media, nports=4, *args, **kw):
        Device.__init__(self, media=media, *args, **kw)
        self.nports = nports
    @property
    def ntwk(self):
        a = self.media.match(nports=4)
        for m,n in [(0,1),(1,0),(2,0),(0,2),(3,2),(2,3)]:
            a.s[:,m,n]=-1j
        for m,n in [(3,1),(1,3)]:
            a.s[:,m,n]=1j
        
        a.s = a.s*1/sqrt(2)
        if self.nports ==3:
            match = self.media.match()
            a = connect(a,3,match,0)
        return a
        
class DualCoupler(Device):
    '''
    Pair of back-to-back directional couplers
    
    Ports are as follows:
        * 0 : insertion of coupler 1
        * 1 : insertion on coupler 2
        * 2 : coupled on coupler 1
        * 3 : coupled on coupler 2
    '''
    def __init__(self, media, c1=1/sqrt(2), c2=None, c1kw={},c2kw ={}):
        Device.__init__(self,media=media)
        if c2 is None:
            c2= c1
        self.c1 = MatchedSymmetricCoupler(media=media,c=c1,nports=3,**c1kw)
        self.c2 = MatchedSymmetricCoupler(media=media,c=c2,nports=3,**c2kw)
        
    @property
    def ntwk(self):
        c1ntwk = self.c1.ntwk
        c2ntwk = self.c2.ntwk
        ntwk = connect(c1ntwk,1,c2ntwk,1)
        ntwk.renumber([0,1,2,3],[0,2,1,3])
        return ntwk



"""
.. module:: skrf.io.touchstone

========================================
touchstone (:mod:`skrf.io.touchstone`)
========================================

Touchstone class

.. autosummary::
   :toctree: generated/

   Touchstone

Functions related to reading/writing touchstones.
-------------------------------------------------

.. autosummary::
   :toctree: generated/

   hfss_touchstone_2_gamma_z0
   hfss_touchstone_2_media
   hfss_touchstone_2_network
   read_zipped_touchstones

"""
import re
import os
import zipfile
import numpy
import numpy as npy

from six.moves import xrange

from ..util import get_fid
from ..network import Network
from ..frequency import Frequency
from ..media import Media, DefinedGammaZ0
from .. import mathFunctions as mf


class Touchstone:
    """
    class to read touchstone s-parameter files

    The reference for writing this class is the draft of the
    Touchstone(R) File Format Specification Rev 2.0 [#]_ and
    Touchstone(R) File Format Specification Version 2.0 [##]_

    .. [#] https://ibis.org/interconnect_wip/touchstone_spec2_draft.pdf
    .. [##] https://ibis.org/touchstone_ver2.0/touchstone_ver2_0.pdf
    """
    def __init__(self, file):
        """
        constructor

        Parameters
        -------------
        file : str or file-object
            touchstone file to load

        Examples
        ---------
        From filename

        >>> t = rf.Touchstone('network.s2p')

        From file-object

        >>> file = open('network.s2p')
        >>> t = rf.Touchstone(file)
        """
        fid = get_fid(file)
        filename = fid.name
        ## file name of the touchstone data file
        self.filename = filename

        ## file format version. 
        # Defined by default to 1.0, since version number can be omitted in V1.0 format
        self.version = '1.0'
        ## comments in the file header
        self.comments = None
        ## unit of the frequency (Hz, kHz, MHz, GHz)
        self.frequency_unit = None
        ## number of frequency points 
        self.frequency_nb = None
        ## s-parameter type (S,Y,Z,G,H)
        self.parameter = None
        ## s-parameter format (MA, DB, RI)
        self.format = None
        ## reference resistance, global setup
        self.resistance = None
        ## reference impedance for each s-parameter
        self.reference = None

        ## numpy array of original s-parameter data
        self.sparameters = None
        ## numpy array of original noise data
        self.noise = None

        ## kind of s-parameter data (s1p, s2p, s3p, s4p)
        self.rank = None
        ## Store port names in a list if they exist in the file
        self.port_names = None

        self.comment_variables=None
        self.load_file(fid)

    def load_file(self, fid):
        """
        Load the touchstone file into the interal data structures
        """

        filename=self.filename

        # Check the filename extension. 
        # Should be .sNp for Touchstone format V1.0, and .ts for V2
        extension = filename.split('.')[-1].lower()
        
        if (extension[0] == 's') and (extension[-1] == 'p'): # sNp
            # check if N is a correct unmber
            try:
                self.rank = int(extension[1:-1])
            except (ValueError):
                raise (ValueError("filename does not have a s-parameter extension. It has  [%s] instead. please, correct the extension to of form: 'sNp', where N is any integer." %(extension)))
        elif extension == 'ts':
            pass
        else:
            raise Exception('Filename does not have the expected Touchstone extension (.sNp or .ts)')

        linenr = 0
        values = []
        while (1):
            linenr +=1
            line = fid.readline()
            if not type(line) == str:
                line = line.decode("ascii")  # for python3 zipfile compatibility
            if not line:
                break

            # store comments if they precede the option line
            line = line.split('!',1)
            if len(line) == 2:
                if not self.parameter:
                    if self.comments == None:
                        self.comments = ''
                    self.comments = self.comments + line[1]
                elif line[1].startswith(' Port['):
                    try:
                        port_string, name = line[1].split('=', 1) #throws ValueError on unpack
                        name = name.strip()
                        garbage, index = port_string.strip().split('[', 1) #throws ValueError on unpack
                        index = int(index.rstrip(']')) #throws ValueError on not int-able
                        if index > self.rank or index <= 0:
                            print("Port name {0} provided for port number {1} but that's out of range for a file with extension s{2}p".format(name, index, self.rank))
                        else:
                            if self.port_names is None: #Initialize the array at the last minute
                                self.port_names = [''] * self.rank
                            self.port_names[index - 1] = name
                    except ValueError as e:
                        print("Error extracting port names from line: {0}".format(line))

            # remove the comment (if any) so rest of line can be processed.
            # touchstone files are case-insensitive
            line = line[0].strip().lower()

            # skip the line if there was nothing except comments
            if len(line) == 0:
                continue

            # grab the [version] string
            if line[:9] == '[version]':
                self.version = line.split()[1]
                continue

            # grab the [reference] string
            if line[:11] == '[reference]':
                # The reference impedances can be span after the keyword  
                # or on the following line
                self.reference = [ float(r) for r in line.split()[2:] ]
                if not self.reference:
                    line = fid.readline()
                    self.reference = [ float(r) for r in line.split()]
                continue
            
            # grab the [Number of Ports] string
            if line[:17] == '[number of ports]':
                self.rank = int(line.split()[-1])
                continue
            
            # grab the [Number of Frequencies] string
            if line[:23] == '[number of frequencies]':
                self.frequency_nb = line.split()[-1]
                continue

            # skip the [Network Data] keyword
            if line[:14] == '[network data]':
                continue
            
            # skip the [End] keyword
            if line[:5] == '[end]':
                continue
            
            # the option line
            if line[0] == '#':
                toks = line[1:].strip().split()
                # fill the option line with the missing defaults
                toks.extend(['ghz', 's', 'ma', 'r', '50'][len(toks):])
                self.frequency_unit = toks[0]
                self.parameter = toks[1]
                self.format = toks[2]
                self.resistance = toks[4]
                if self.frequency_unit not in ['hz', 'khz', 'mhz', 'ghz']:
                    print('ERROR: illegal frequency_unit [%s]',  self.frequency_unit)
                    # TODO: Raise
                if self.parameter not in 'syzgh':
                    print('ERROR: illegal parameter value [%s]', self.parameter)
                    # TODO: Raise
                if self.format not in ['ma', 'db', 'ri']:
                    print('ERROR: illegal format value [%s]', self.format)
                    # TODO: Raise

                continue

            # collect all values without taking care of there meaning
            # we're seperating them later
            values.extend([ float(v) for v in line.split() ])

        # let's do some post-processing to the read values
        # for s2p parameters there may be noise parameters in the value list
        values = numpy.asarray(values)
        if self.rank == 2:
            # the first frequency value that is smaller than the last one is the
            # indicator for the start of the noise section
            # each set of the s-parameter section is 9 values long
            pos = numpy.where(numpy.sign(numpy.diff(values[::9])) == -1)
            if len(pos[0]) != 0:
                # we have noise data in the values
                pos = pos[0][0] + 1   # add 1 because diff reduced it by 1
                noise_values = values[pos*9:]
                values = values[:pos*9]
                self.noise = noise_values.reshape((-1,5))

        if len(values)%(1+2*(self.rank)**2) != 0 :
            # incomplete data line / matrix found
            raise AssertionError

        # reshape the values to match the rank
        self.sparameters = values.reshape((-1, 1 + 2*self.rank**2))
        # multiplier from the frequency unit
        self.frequency_mult = {'hz':1.0, 'khz':1e3,
                               'mhz':1e6, 'ghz':1e9}.get(self.frequency_unit)
        # set the reference to the resistance value if no [reference] is provided
        if not self.reference:
            self.reference = [self.resistance] * self.rank

    def get_comments(self, ignored_comments = ['Created with skrf']):
        """
        Returns the comments which appear anywhere in the file.  Comment lines
        containing ignored comments are removed.  By default these are comments
        which contain special meaning withing skrf and are not user comments.
        """
        processed_comments = ''
        if self.comments is None:
            self.comments = ''
        for comment_line in self.comments.split('\n'):
            for ignored_comment in ignored_comments:
                if ignored_comment in comment_line:
                        comment_line = None
            if comment_line:
                processed_comments = processed_comments + comment_line + '\n'
        return processed_comments
    
    def get_comment_variables(self):
        '''
        convert hfss variable comments to a dict of vars:(numbers,units)
        '''
        comments = self.comments
        p1 = re.compile('\w* = \w*')
        p2 = re.compile('\s*(\d*)\s*(\w*)')
        var_dict = {}
        for k in re.findall(p1, comments):
            var, value = k.split('=')
            var=var.rstrip()
            try:
                var_dict[var] = p2.match(value).groups()
            except:
                pass
        return var_dict
    
    def get_format(self, format="ri"):
        """
        returns the file format string used for the given format.
        This is useful to get some information.
        """
        if format == 'orig':
            frequency = self.frequency_unit
            format = self.format
        else:
            frequency = 'hz'
        return "%s %s %s r %s" %(frequency, self.parameter,
                                 format, self.resistance)


    def get_sparameter_names(self, format="ri"):
        """
        Generate a list of column names for the s-parameter data. 
        The names are different for each format.

        Parameters
        ----------
        format : str
          Format: ri, ma, db, orig (where orig refers to one of the three others)

        Returns
        -------
        names : list
            list of strings

        """
        names = ['frequency']
        if format == 'orig':
            format = self.format
        ext1, ext2 = {'ri':('R','I'),'ma':('M','A'), 'db':('DB','A')}.get(format)
        for r1 in xrange(self.rank):
            for r2 in xrange(self.rank):
                names.append("S%i%i%s"%(r1+1,r2+1,ext1))
                names.append("S%i%i%s"%(r1+1,r2+1,ext2))
        return names

    def get_sparameter_data(self, format='ri'):
        """
        get the data of the s-parameter with the given format.

        Parameters
        ----------
        format : str
          Format: ri, ma, db, orig

        supported formats are:
          orig:  unmodified s-parameter data
          ri:    data in real/imaginary
          ma:    data in magnitude and angle (degree)
          db:    data in log magnitute and angle (degree)

        Returns
        -------
        ret: list
            list of numpy.arrays

        """
        ret = {}
        if format == 'orig':
            values = self.sparameters
        else:
            values = self.sparameters.copy()
            # use frequency in hz unit
            values[:,0] = values[:,0]*self.frequency_mult
            if (self.format == 'db') and (format == 'ma'):
                values[:,1::2] = 10**(values[:,1::2]/20.0)
            elif (self.format == 'db') and (format == 'ri'):
                v_complex = ((10**values[:,1::2]/20.0)
                             * numpy.exp(1j*numpy.pi/180 * values[:,2::2]))
                values[:,1::2] = numpy.real(v_complex)
                values[:,2::2] = numpy.imag(v_complex)
            elif (self.format == 'ma') and (format == 'db'):
                values[:,1::2] = 20*numpy.log10(values[:,1::2])
            elif (self.format == 'ma') and (format == 'ri'):
                v_complex = (values[:,1::2] * numpy.exp(1j*numpy.pi/180 * values[:,2::2]))
                values[:,1::2] = numpy.real(v_complex)
                values[:,2::2] = numpy.imag(v_complex)
            elif (self.format == 'ri') and (format == 'ma'):
                v_complex = numpy.absolute(values[:,1::2] + 1j* self.sparameters[:,2::2])
                values[:,1::2] = numpy.absolute(v_complex)
                values[:,2::2] = numpy.angle(v_complex)*(180/numpy.pi)
            elif (self.format == 'ri') and (format == 'db'):
                v_complex = numpy.absolute(values[:,1::2] + 1j* self.sparameters[:,2::2])
                values[:,1::2] = 20*numpy.log10(numpy.absolute(v_complex))
                values[:,2::2] = numpy.angle(v_complex)*(180/numpy.pi)

        for i,n in enumerate(self.get_sparameter_names(format=format)):
            ret[n] = values[:,i]

        # transpose Touchstone V1 2-port files (.2p), as the order is (11) (21) (12) (22)
        file_name_ending = self.filename.split('.')[-1].lower()
        if self.rank == 2 and file_name_ending == "s2p":
            swaps = [ k for k in ret if '21' in k]
            for s in swaps:
                true_s = s.replace('21', '12')
                ret[s], ret[true_s] = ret[true_s], ret[s]

        return ret

    def get_sparameter_arrays(self):
        """
        Returns the s-parameters as a tuple of arrays.

        The first element is the frequency vector (in Hz) and the s-parameters are a 3d numpy array.
        The values of the s-parameters are complex number.

        Example
        -------
          f,a = self.sgetparameter_arrays()
          s11 = a[:,0,0]

        """
        v = self.sparameters

        if self.format == 'ri':
            v_complex = v[:,1::2] + 1j* v[:,2::2]
        elif self.format == 'ma':
            v_complex = (v[:,1::2] * numpy.exp(1j*numpy.pi/180 * v[:,2::2]))
        elif self.format == 'db':
            v_complex = ((10**(v[:,1::2]/20.0)) * numpy.exp(1j*numpy.pi/180 * v[:,2::2]))

        if self.rank == 2 :
            # this return is tricky; it handles the way touchtone lines are
            # in case of rank==2: order is s11,s21,s12,s22
            return (v[:,0] * self.frequency_mult,
                    numpy.transpose(v_complex.reshape((-1, self.rank, self.rank)),axes=(0,2,1)))
        else:
            return (v[:,0] * self.frequency_mult,
                    v_complex.reshape((-1, self.rank, self.rank)))

    def get_noise_names(self):
        """
        TODO: NIY
        """
        TBD = 1


    def get_noise_data(self):
        """
        TODO: NIY
        """
        TBD = 1
        noise_frequency = noise_values[:,0]
        noise_minimum_figure = noise_values[:,1]
        noise_source_reflection = noise_values[:,2]
        noise_source_phase = noise_values[:,3]
        noise_normalized_resistance = noise_values[:,4]
        
    def is_from_hfss(self):
        '''
        Check if the Touchstone file has been produced by HFSS
        
        Returns
        ------------
        status : boolean
            True if the Touchstone file has been produced by HFSS
            False otherwise
        '''
        if self.comments is None:
            return False

        if 'exported from hfss' in str.lower(self.comments):
            return True

        return False

    def get_gamma_z0(self):
        '''
        Extracts Z0 and Gamma comments from touchstone file (is provided)
        
        Returns
        --------
        gamma : complex numpy.ndarray
            complex  propagation constant
        z0 : numpy.ndarray
            complex port impedance    
        '''
        def line2ComplexVector(s):
            return mf.scalar2Complex(npy.array([k for k in s.strip().split(' ')
                                                if k != ''][self.rank*-2:],
                                                dtype='float'))
    
        with open(self.filename) as f:
            gamma, z0 = [],[]
            
            for line in f:

                # HFSS adds gamma and z0 data in .sNp files using comments.
                # NB : each line(s) describe gamma and z0. 
                #  But, depending on the HFSS version, either:
                #  - up to 4 ports only. 
                #        for N > 4, gamma and z0 are given by additional lines
                #  - all gamma and z0 are given on a single line (since 2020R2)
                # In addition, some spurious '!' can remain in these lines
                if '! Gamma' in line:
                    _line = line.replace('! Gamma', '').replace('!', '').rstrip()

                    # check how many elements are in the first line
                    nb_elem = len(_line.split())

                    if nb_elem == 2*self.rank:
                        # case of all data in a single line
                        gamma.append(line2ComplexVector(_line.replace('!', '').rstrip()))
                    else:
                        # case of Nport > 4 *and* data on additional multiple lines
                        for _ in range(int(npy.ceil(self.rank/4.0)) - 1):
                            _line += next(f).replace('!', '').rstrip()
                        gamma.append(line2ComplexVector(_line))

             
                if '! Port Impedance' in line:
                    _line = line.replace('! Port Impedance', '').rstrip()
                    nb_elem = len(_line.split())

                    if nb_elem == 2*self.rank:
                        z0.append(line2ComplexVector(_line.replace('!', '').rstrip()))
                    else:
                        for _ in range(int(npy.ceil(self.rank/4.0)) - 1):
                            _line += next(f).replace('!', '').rstrip()
                        z0.append(line2ComplexVector(_line))
    
            # If the file does not contain valid port impedance comments, set to default one
            if len(z0) == 0:
                z0 = npy.complex(self.resistance)
                #raise ValueError('Touchstone does not contain valid gamma, port impedance comments')


        return npy.array(gamma), npy.array(z0)

def hfss_touchstone_2_gamma_z0(filename):
    '''
    Extracts Z0 and Gamma comments from touchstone file

    Takes a HFSS-style touchstone file with Gamma and Z0 comments and
    extracts a triplet of arrays being: (frequency, Gamma, Z0)

    Parameters
    ------------
    filename : string
        the HFSS-style touchstone file


    Returns
    --------
    f : numpy.ndarray
        frequency vector (in Hz)
    gamma : complex numpy.ndarray
        complex  propagation constant
    z0 : numpy.ndarray
        complex port impedance

    Examples
    ----------
    >>> f,gamm,z0 = rf.hfss_touchstone_2_gamma_z0('line.s2p')
    '''
    ntwk = Network(filename)

    return ntwk.frequency.f, ntwk.gamma, ntwk.z0


def hfss_touchstone_2_media(filename, f_unit='ghz'):
    '''
    Creates a :class:`~skrf.media.Media` object from a a HFSS-style touchstone file with Gamma and Z0 comments

    Parameters
    ------------
    filename : string
        the HFSS-style touchstone file
    f_unit : ['hz','khz','mhz','ghz']
        passed to f_unit parameters of Frequency constructor

    Returns
    --------
    my_media : skrf.media.Media object
        the transmission line model defined by the gamma, and z0
        comments in the HFSS file.

    Examples
    ----------
    >>> port1_media, port2_media = rf.hfss_touchstone_2_media('line.s2p')

    See Also
    ---------
    hfss_touchstone_2_gamma_z0 : returns gamma, and z0
    '''
    ntwk = Network(filename)

    freq = ntwk.frequency
    gamma = ntwk.gamma
    z0 = ntwk.z0
    

    media_list = []

    for port_n in range(gamma.shape[1]):
        media_list.append(\
            DefinedGammaZ0(
                frequency = freq,
                gamma =  gamma[:, port_n],
                z0 = z0[:, port_n]
                )
            )


    return media_list


def hfss_touchstone_2_network(filename, f_unit='ghz'):
    '''
    Creates a :class:`~skrf.Network` object from a a HFSS-style touchstone file

    Parameters
    ------------
    filename : string
        the HFSS-style touchstone file
    f_unit : ['hz','khz','mhz','ghz']
        passed to f_unit parameters of Frequency constructor

    Returns
    --------
    my_network : skrf.Network object
        the n-port network model

    Examples
    ----------
    >>> my_network = rf.hfss_touchstone_2_network('DUT.s2p')

    See Also
    ---------
    hfss_touchstone_2_gamma_z0 : returns gamma, and z0
    '''
    my_network = Network(file=filename, f_unit=f_unit)
    return(my_network)


def read_zipped_touchstones(ziparchive, dir=""):
    """
    similar to skrf.io.read_all_networks, which works for directories but only for touchstones in ziparchives

    Parameters
    ----------
    ziparchive : zipfile.ZipFile
        an zip archive file, containing touchstone files and open for reading
    dir : str
        the directory of the ziparchive to read networks from, default is "" which reads only the root directory

    Returns
    -------
    dict
    """
    networks = dict()
    for fname in ziparchive.namelist():  # type: str
        directory, filename = os.path.split(fname)
        if dir == directory and fname[-4:].lower() in (".s1p", ".s2p", ".s3p", ".s4p"):
            network = Network.zipped_touchstone(fname, ziparchive)
            networks[network.name] = network
    return networks



import warnings
from collections import OrderedDict, Iterable

import numpy as np
import skrf
import pyvisa

from . import abcvna
from . import keysight_fieldfox_scpi


class FieldFox(abcvna.VNA):
    """
    Class for Keysight FieldFox Signal Analyzers
    """

    DEFAULT_VISA_ADDRESS = "USB0::0x2A8D::0x5C18::MY56071097::INSTR"
    NAME = "Keysight FieldFox"
    NPORTS = 2
    NCHANNELS = 1
    SCPI_VERSION_TESTED = 'A.06.23'

    def __init__(self, address=DEFAULT_VISA_ADDRESS, **kwargs):
        """
        initialization of FieldFox Class

        Parameters
        ----------
        address : str
            visa resource string (full string or ip address)
        kwargs : dict
            interface (str), port (int), timeout (int),
        :param address:
        :param kwargs:
        """
        super(FieldFox, self).__init__(address, **kwargs)
        self.resource.timeout = kwargs.get("timeout", 2000)
        self.scpi = keysight_fieldfox_scpi.SCPI(self.resource)
        self.use_ascii()

    def use_binary(self):
        raise Exception("binary not working correctly")
        """setup the analyzer to transfer in binary which is faster, especially for large datasets"""
        self.resource.write(':FORM:BORD SWAP')
        self.resource.write(':FORM:DATA REAL,64')
        # self.resource.write(':FORM:DATA REAL')
        self.resource.values_format.use_binary(datatype='d', is_big_endian=False, container=np.array)

    def use_ascii(self):
        self.resource.write(':FORM:DATA ASCII')
        self.resource.values_format.use_ascii(converter='f', separator=',', container=np.array)

    @property
    def echo(self):
        return self.scpi.echo

    @echo.setter
    def echo(self, onoff):
        if onoff in (1, True):
            self.scpi.echo = True
        elif onoff in (0, False):
            self.scpi.echo = False
        else:
            raise warnings.warn("echo must be a boolean")

    @property
    def n_traces(self):
        return self.scpi.query_trace_count()

    def sweep(self, **kwargs):
        """
        Initialize a fresh sweep of data on the specified channels

        Parameters
        ----------
        kwargs : dict
            channel ("all", int or list of channels), timeout (milliseconds)

        trigger a fresh sweep on the specified channels (default is "all" if no channel specified)
        Autoset timeout and sweep mode based upon the analyzers current averaging setting,
        and then return to the prior state of continuous trigger or hold.
        """
        self.resource.clear()
        original_timeout = self.resource.timeout
        self.resource.timeout = kwargs.get("timeout", 15000)  # allow at least 10 seconds for every averaging sweep

        was_continuous = self.scpi.query_continuous_sweep()
        self.scpi.set_continuous_sweep(False)
        self.scpi.set_clear_averaging()

        try:
            for i in range(self.scpi.query_averaging_count()):
                self.scpi.query(":INIT;*OPC?")
        finally:
            self.resource.timeout = original_timeout
            self.scpi.set_continuous_sweep(was_continuous)

    def upload_twoport_calibration(self, cal, port1=1, port2=2, **kwargs):
        """
        upload a calibration to the vna, and set correction on all measurements

        Parameters
        ----------
        cal : skrf.Calibration
        port1: int
        port2: int

        calibration error terms reference
            # forward = (1, 1), reverse = (2, 2)
            "directivity": "EDIR",
            "source match": "ESRM",
            "reflection tracking": "ERFT",

            # forward = (2, 1), reverse = (1, 2)
            "load match": "ELDM",
            "transmission tracking": "ETRT"
            "isolation": "EXTLK"

        """
        self.active_channel = channel = kwargs.get("channel", self.active_channel)

        calname = kwargs.get("calname", "skrf_12term_cal")
        calibrations = self.scpi.query_calset_catalog(cnum=channel, form="NAME")
        if calname in calibrations:
            self.scpi.set_delete_calset(cnum=channel, calset_name=calname)
        self.scpi.set_create_calset(cnum=channel, calset_name=calname)

        cfs = dict()
        for coef, data in cal.coefs_12term.items():
            cfs[coef] = skrf.mf.complex2Scalar(data)

        for eterm, coef in zip(("EDIR", "ESRM", "ERFT"), ("directivity", "source match", "reflection tracking")):
            self.scpi.set_calset_data(channel, eterm, port1, port1, eterm_data=cfs["forward " + coef])
            self.scpi.set_calset_data(channel, eterm, port2, port2, eterm_data=cfs["reverse " + coef])
        for eterm, coef in zip(("ELDM", "ETRT", "EXTLK"), ("load match", "transmission tracking", "isolation")):
            self.scpi.set_calset_data(channel, eterm, port2, port1, eterm_data=cfs["forward " + coef])
            self.scpi.set_calset_data(channel, eterm, port1, port2, eterm_data=cfs["reverse " + coef])

        self.scpi.set_active_calset(1, calname, True)

    def initiate_oneport(self, port=1, **kwargs):
        self.scpi.set_instrument("NA")
        self.scpi.set_trace_display_config("D1")
        if port == 1:
            self.scpi.set_trace_measurement(1, "S11")
        elif port == 2:
            self.scpi.set_trace_measurement(1, "S22")
        else:
            raise Exception("wrong port, must be 1 or 2")
        self.wait_until_finished()
        self.autoscale_all()

    def initiate_twoport(self):
        self.scpi.set_instrument("NA")
        self.scpi.set_trace_display_config("D12_34")
        self.scpi.set_trace_measurement(1, "S11")
        self.scpi.set_trace_measurement(2, "S21")
        self.scpi.set_trace_measurement(3, "S12")
        self.scpi.set_trace_measurement(4, "S22")
        self.wait_until_finished()
        self.autoscale_all()

    def autoscale_all(self):
        for tr in range(1, self.n_traces + 1):
            self.scpi.set_trace_autoscale(tr)

    def get_list_of_traces(self, **kwargs):
        return

        # self.resource.clear()
        # traces = []
        # channels = self.scpi.query_available_channels()
        # for channel in channels:
        #     meas_list = self.scpi.query_meas_name_list(channel)
        #     if len(meas_list) == 1:
        #         continue  # if there isnt a single comma, then there aren't any measurments
        #     parameters = dict([(meas_list[k], meas_list[k + 1]) for k in range(0, len(meas_list) - 1, 2)])
        #
        #     meas_numbers = self.scpi.query_meas_number_list()
        #     for mnum in meas_numbers:
        #         name = self.scpi.query_meas_name_from_number(mnum)
        #         item = {"name": name, "channel": channel, "measurement number": mnum,
        #                 "parameter": parameters.get(name, name)}
        #         item["label"] = "{:s} - Chan{:},Meas{:}".format(
        #             item["parameter"], item["channel"], item["measurement number"])
        #         traces.append(item)
        # return traces

    def get_traces(self, traces, **kwargs):
        """
        retrieve traces as 1-port networks from a list returned by get_list_of_traces

        Parameters
        ----------
        traces : list
            list of type that is exported by self.get_list_of_traces
        kwargs : dict
            sweep (bool), name_prefix (str)

        Returns
        -------
        list
            a list of 1-port networks representing the desired traces

        Notes
        -----
        There is no current way to distinguish between traces and 1-port networks within skrf
        """
        self.resource.clear()
        sweep = kwargs.get("sweep", False)

        name_prefix = kwargs.get("name_prefix", "")
        if name_prefix:
            name_prefix += " - "

        channels = OrderedDict()
        for trace in traces:
            ch = trace["channel"]
            if ch not in channels.keys():
                channels[ch] = {
                    "frequency": None,
                    "traces": list()}
            channels[ch]["traces"].append(trace)

        if sweep is True:
            self.sweep(channels=list(channels.keys()))

        traces = []
        for ch, ch_data in channels.items():
            frequency = ch_data["frequency"] = self.get_frequency()
            for trace in ch_data["traces"]:
                self.scpi.set_selected_meas_by_number(trace["channel"], trace["measurement number"])
                sdata = self.scpi.query_data(trace["channel"], "SDATA")
                s = sdata[::2] + 1j * sdata[1::2]
                ntwk = skrf.Network()
                ntwk.s = s
                ntwk.frequency = frequency
                ntwk.name = name_prefix + trace.get("parameter", "trace")
                traces.append(ntwk)
        return traces

    def get_frequency(self, **kwargs):
        """
        get an skrf.Frequency object for the current channel

        Parameters
        ----------
        kwargs : dict
            channel (int), f_unit (str)

        Returns
        -------
        skrf.Frequency
        """
        self.resource.clear()
        f_start = self.scpi.query_f_start()
        f_stop = self.scpi.query_f_stop()
        f_npoints = self.scpi.query_sweep_n_points()
        freq = np.linspace(f_start, f_stop, f_npoints)

        frequency = skrf.Frequency.from_f(freq, unit="Hz")
        frequency.unit = kwargs.get("f_unit", "Hz")
        return frequency

    def set_frequency_sweep(self, f_start, f_stop, f_npoints, **kwargs):
        f_unit = kwargs.get("f_unit", "hz").lower()
        if f_unit != "hz":
            f_start = self.to_hz(f_start, f_unit)
            f_stop = self.to_hz(f_stop, f_unit)
        self.scpi.set_f_start(f_start)
        self.scpi.set_f_stop(f_stop)
        self.scpi.set_sweep_n_points(f_npoints)

    def get_twoport(self, ports=(1, 2), **kwargs):
        if not self.scpi.query_trace_display_config() == "D12_34":
            self.initiate_twoport()

        port1, port2 = ports

        ntwk = skrf.Network()
        ntwk.frequency = self.get_frequency(f_unit="GHz")
        npoints = len(ntwk.f)
        ntwk.s = np.empty(shape=(npoints, 2, 2), dtype=complex)

        if kwargs.get("sweep", False) is True:
            self.sweep()

        sdata = []
        for i in range(1,5):
            self.scpi.set_current_trace(i)
            sdata.append(self.scpi.query_current_trace_data())

        ntwk.s[:, 0, 0] = sdata[0][::2] + 1j * sdata[0][1::2]
        ntwk.s[:, 1, 0] = sdata[1][::2] + 1j * sdata[1][1::2]
        ntwk.s[:, 0, 1] = sdata[2][::2] + 1j * sdata[2][1::2]
        ntwk.s[:, 1, 1] = sdata[3][::2] + 1j * sdata[3][1::2]

        ntwk.z0 = kwargs.get("z0", 50.)

        if port1 == 1 and port2 == 2:
            return ntwk
        elif port1 == 2 and port2 ==1:
            ntwk.flip()
            return ntwk
        else:
            raise Exception("ports must be a 2-length tuple with either 1, 2 or 2, 1")

    def get_forward_switch_terms(self, **kwargs):
        print("make sure source port is set to port2 manually")

        self.resource.clear()
        trace_config = self.scpi.query_trace_display_config()
        trace_measurements = []
        for i in range(self.scpi.query_trace_count()):
            trace_measurements.append(self.scpi.query_trace_measurement(i + 1))
        self.scpi.set_trace_count(2)
        self.scpi.set_trace_measurement(1, "R1")
        self.scpi.set_trace_measurement(2, "A")

        self.sweep(**kwargs)

        self.scpi.set_current_trace(1)
        R1 = self.get_active_trace_as_network()
        self.scpi.set_current_trace(2)
        A = self.get_active_trace_as_network()

        self.scpi.set_trace_display_config(trace_config)
        for i in range(len(trace_measurements)):
            self.scpi.set_trace_measurement(i + 1, trace_measurements[i])

        return R1 / A

    def get_reverse_switch_terms(self, **kwargs):
        print("make sure source port is set to port1 manually")

        self.resource.clear()
        trace_config = self.scpi.query_trace_display_config()
        trace_measurements = []
        for i in range(self.scpi.query_trace_count()):
            trace_measurements.append(self.scpi.query_trace_measurement(i + 1))
        self.scpi.set_trace_count(2)
        self.scpi.set_trace_measurement(1, "R2")
        self.scpi.set_trace_measurement(2, "B")

        self.sweep(**kwargs)

        self.scpi.set_current_trace(1)
        R2 = self.get_active_trace_as_network()
        self.scpi.set_current_trace(2)
        B = self.get_active_trace_as_network()

        self.scpi.set_trace_display_config(trace_config)
        for i in range(len(trace_measurements)):
            self.scpi.set_trace_measurement(i + 1, trace_measurements[i])

        return R2 / B

    def display_switch_terms(self):
        self.scpi.set_trace_count(4)
        self.scpi.set_trace_measurement(1, "R1")
        self.scpi.set_trace_measurement(2, "R2")
        self.scpi.set_trace_measurement(3, "A")
        self.scpi.set_trace_measurement(4, "B")

    def set_scale_all(self, bottom, top):
        for i in range(1, self.scpi.query_trace_count() + 1):
            self.write("DISP:WIND:TRAC{:d}:Y:TOP {:}".format(i, top))
            self.write("DISP:WIND:TRAC{:d}:Y:BOTT {:}".format(i, bottom))

    def get_switch_terms(self, **kwargs):
        msg = """the Field Fox VNA does not allow the automatic setting of source ports and so one cannot automatically
        measure the switch terms.  Instead, select Measure->More->Advanced to set the Source Port to port2, then use
        FieldFox.get_forward_switch_terms method to collect the forward switch terms.  Then change the source port to
        to port1, then use FieldFox.get_reverse_switch_terms to collect the reverse switch terms"""
        raise Exception(msg)

    def get_trace(self, tnum):
        self.scpi.set_current_trace(tnum)
        return self.get_active_trace_as_network()

    def get_active_trace_as_network(self, **kwargs):
        """
        get the active trace as a network object

        Parameters
        ----------
        kwargs : dict
            channel (int), sweep (bool)

        Returns
        -------
        skrf.Network
        """
        sweep = kwargs.get("sweep", False)
        if sweep:
            self.sweep()

        ntwk = skrf.Network()
        sdata = self.scpi.query_current_trace_data()
        ntwk.s = sdata[::2] + 1j * sdata[1::2]
        ntwk.frequency = self.get_frequency()
        return ntwk



import unittest
import os
import warnings
import six.moves.cPickle as pickle
import skrf as rf
import numpy as npy
from numpy.random  import rand, uniform
from nose.tools import nottest
from nose.plugins.skip import SkipTest

from skrf.calibration import OnePort, PHN, SDDL, TRL, SOLT, UnknownThru, EightTerm, TwoPortOnePath, EnhancedResponse,TwelveTerm, SixteenTerm, LMR16, terminate, determine_line, determine_reflect, NISTMultilineTRL

from skrf import two_port_reflect
from skrf.networkSet import NetworkSet
from skrf.util import suppress_warning_decorator
from skrf.media import DistributedCircuit

# number of frequency points to test calibration at .
# i choose 1 for speed, but given that many tests employ *random* 
# networks values >100 are better for  initialy verification
global NPTS  
NPTS = 1

WG_lossless =  rf.RectangularWaveguide(rf.F(75,100,NPTS), a=100*rf.mil, z0=50)
WG =  rf.RectangularWaveguide(rf.F(75,100,NPTS), a=100*rf.mil, z0=50, rho='gold')


class DetermineTest(unittest.TestCase):
    def setUp(self):
        self.wg = WG
        wg = self.wg
        self.X = wg.random(n_ports =2, name = 'X')
        self.Y = wg.random(n_ports =2, name='Y')  
        
        
        #thru
        self.T = wg.thru()
        self.T_m = self.embed(self.T)
        
        #line
        self.L = wg.line(80,'deg')
        self.L_approx = self.wg.line(90,'deg')
        self.L_m = self.embed(self.L)
        
        # reflect
        self.r = [ wg.delay_load(p,k,'deg') \
                    for k in [-10,10,88,92] \
                    for p in [-.9,-1]]
        
        
        self.R = [rf.two_port_reflect(k,k) for k in self.r]
        
        short= wg.short()
        open = wg.open()
        self.r_estimate = [short, short,short, short, open ,open, open ,open]
        self.R_m = [self.embed(k) for k in self.R]

    def embed(self,x):
        return self.X**x**self.Y    
        
    def test_determine_line(self):
        L_found = determine_line(self.T_m, self.L_m, 
                                 line_approx=self.L_approx)
        self.assertEqual(L_found,self.L)

    def test_determine_reflect(self):
        r_found = [determine_reflect(self.T_m,k,self.L_m, l, self.L_approx) \
                   for k,l in zip(self.R_m, self.r_estimate)]

        [ self.assertEqual(k,l) for k,l in zip(self.r, r_found)]


class CalibrationTest(object):
    '''
    This is the generic Calibration test case which all Calibration 
    Subclasses should be able to pass. They must implement
    '''

    def test_accuracy_of_dut_correction(self):
        a = self.wg.random(n_ports=self.n_ports, name = 'actual')
        m = self.measure(a)
        c = self.cal.apply_cal(m)
        c.name = 'corrected'   
        self.assertEqual(c,a)
        
    def test_error_ntwk(self):
        a= self.cal.error_ntwk 
    
    def test_coefs_ntwks(self):
        a= self.cal.coefs_ntwks
    
    @suppress_warning_decorator("only gave a single measurement orientation")
    def test_caled_ntwks(self):
        a= self.cal.caled_ntwks
        
    @suppress_warning_decorator("only gave a single measurement orientation")
    def test_residual_ntwks(self):
        a= self.cal.residual_ntwks
    
    def test_embed_then_apply_cal(self):
        a = self.wg.random(n_ports=self.n_ports)
        self.assertEqual(self.cal.apply_cal(self.cal.embed(a)),a)

    def test_embed_equal_measure(self):
        a = self.wg.random(n_ports=self.n_ports)
        self.assertEqual(self.cal.embed(a),self.measure(a))
        
    @suppress_warning_decorator("n_thrus is None")
    def test_from_coefs(self):
        cal_from_coefs = self.cal.from_coefs(self.cal.frequency, self.cal.coefs)
        ntwk = self.wg.random(n_ports=self.n_ports)
        self.assertEqual(cal_from_coefs.apply_cal(self.cal.embed(ntwk)),ntwk)
        
    @suppress_warning_decorator("n_thrus is None")
    def test_from_coefs_ntwks(self):
        cal_from_coefs = self.cal.from_coefs_ntwks(self.cal.coefs_ntwks)
        ntwk = self.wg.random(n_ports=self.n_ports)
        self.assertEqual(cal_from_coefs.apply_cal(self.cal.embed(ntwk)),ntwk)


class OnePortTest(unittest.TestCase, CalibrationTest):
    '''
    One-port calibration test.


    '''
    def setUp(self):
        self.n_ports = 1
        self.wg = WG
        wg = self.wg
                
        self.E = wg.random(n_ports =2, name = 'E')
        
        ideals = [
                wg.short( name='short'),
                wg.delay_short( 45.,'deg',name='ew'),
                wg.delay_short( 90.,'deg',name='qw'),
                wg.match( name='load'),
                ]
        measured = [self.measure(k) for k in ideals]
        
        self.cal = rf.OnePort(
            is_reciprocal = True, 
            ideals = ideals, 
            measured = measured,
            )
        
    def measure(self, ntwk):
        out = self.E**ntwk
        out.name = ntwk.name
        return out

    def test_accuracy_of_directivity(self):
        self.assertEqual(
            self.E.s11, 
            self.cal.coefs_ntwks['directivity'],
            )
        
    def test_accuracy_of_source_match(self):
        self.assertEqual(
            self.E.s22, 
            self.cal.coefs_ntwks['source match'],
            )
    
    def test_accuracy_of_reflection_tracking(self):
        self.assertEqual(
            self.E.s21*self.E.s12, 
            self.cal.coefs_ntwks['reflection tracking'],
            )


class SDDLTest(OnePortTest):
    def setUp(self):
        #raise SkipTest('Doesnt work yet')
        self.n_ports = 1
        #Exact only with a lossless waveguide
        self.wg = WG_lossless
        wg = self.wg
        
        self.E = wg.random(n_ports =2, name = 'E')
        #self.E.s[0,:,:] = npy.array([[.1j,1],[1j,1j+2]])
        #print self.E.s[0]
        
        ideals = [
                wg.short( name='short'),
                wg.delay_short( 45.,'deg',name='ew'),
                wg.delay_short( 90.,'deg',name='qw'),
                wg.load(.2+.2j, name='load'),
                ]
        actuals = [
                wg.short( name='short'),
                wg.delay_short( 10.,'deg',name='ew'),
                wg.delay_short( 33.,'deg',name='qw'),
                wg.load(.2+.2j, name='load'),
                ]
        measured = [self.measure(k) for k in actuals]
        
        self.cal = rf.SDDL(
            is_reciprocal = True, 
            ideals = ideals, 
            measured = measured,
            )
    
    def test_init_with_nones(self):
        wg=self.wg
        wg.frequency = rf.F.from_f([100])
        
        self.E = wg.random(n_ports =2, name = 'E')
        
        ideals = [
                wg.short( name='short'),
                None, 
                None,
                wg.load(.2+.2j, name='load'),
                ]
        actuals = [
                wg.short( name='short'),
                wg.delay_short( 10.,'deg',name='ew'),
                wg.delay_short( 33.,'deg',name='qw'),
                wg.load(.2+.2j, name='load'),
                ]
        measured = [self.measure(k) for k in actuals]
        
        self.cal = rf.SDDL(
            is_reciprocal = True, 
            ideals = ideals, 
            measured = measured,
            )
        self.cal.run()
    
    def test_from_coefs(self):
        raise SkipTest('not applicable ')
    def test_from_coefs_ntwks(self):
        raise SkipTest('not applicable ')


class SDDLWeikle(OnePortTest):
    def setUp(self):
        #raise SkipTest('Doesnt work yet')
        self.n_ports = 1
        #Exact only with a lossless waveguide
        self.wg = WG_lossless
        wg = self.wg
        self.E = wg.random(n_ports =2, name = 'E')
        #self.E.s[0,:,:] = npy.array([[.1j,1],[1j,1j+2]])
        #print self.E.s[0]
        
        ideals = [
                wg.short( name='short'),
                wg.delay_short( 45.,'deg',name='ew'),
                wg.delay_short( 90.,'deg',name='qw'),
                wg.load(.2+.2j, name='load'),
                ]
        actuals = [
                wg.short( name='short'),
                wg.delay_short( 10.,'deg',name='ew'),
                wg.delay_short( 80.,'deg',name='qw'),
                wg.load(.2+.2j, name='load'),
                ]
        measured = [self.measure(k) for k in actuals]
        
        self.cal = rf.SDDLWeikle(
            is_reciprocal = True, 
            ideals = ideals, 
            measured = measured,
            )
    
    def test_from_coefs(self):
        raise SkipTest('not applicable ')
    def test_from_coefs_ntwks(self):
        raise SkipTest('not applicable ')


class SDDMTest(OnePortTest):
    '''
    This is a specific test of SDDL to verify it works when the load is 
    a matched load. This test has been used to show that the SDDLWeikle 
    variant fails, with a perfect matched load. 
    '''
    def setUp(self):
        
        self.n_ports = 1
        #Exact only with a lossless waveguide
        self.wg = WG_lossless
        wg = self.wg
        
        self.E = wg.random(n_ports =2, name = 'E')
        
        ideals = [
                wg.short( name='short'),
                wg.delay_short( 45.,'deg',name='ew'),
                wg.delay_short( 90.,'deg',name='qw'),
                wg.match( name='load'),
                ]
        actuals = [
                wg.short( name='short'),
                wg.delay_short( 10.,'deg',name='ew'),
                wg.delay_short( 80.,'deg',name='qw'),
                wg.match(name='load'),
                ]
        measured = [self.measure(k) for k in actuals]
        
        self.cal = rf.SDDL(
            is_reciprocal = True, 
            ideals = ideals, 
            measured = measured,
            )
    
    def test_from_coefs(self):
        raise SkipTest('not applicable ')
    
    def test_from_coefs_ntwks(self):
        raise SkipTest('not applicable ')


@SkipTest
class PHNTest(OnePortTest):
    '''
    '''
    def setUp(self):
        
        self.n_ports = 1
        self.wg = WG
        wg = self.wg
        
        self.E = wg.random(n_ports =2, name = 'E')
        known1 = wg.random()
        known2 = wg.random()
        #known1 = wg.short()
        #known2 = wg.load(rand() + rand()*1j) 
        
        ideals = [
                wg.delay_short( 45.,'deg',name='ideal ew'),
                wg.delay_short( 90.,'deg',name='ideal qw'),
                known1,
                known2,
                ]
        actuals = [
                wg.delay_short( 33.,'deg',name='true ew'),
                wg.delay_short( 110.,'deg',name='true qw'),
                known1,
                known2,
                ]
        measured = [self.measure(k) for k in actuals]
        self.actuals = actuals 
        
        self.cal = PHN(
            is_reciprocal = True, 
            ideals = ideals, 
            measured = measured,
            )
       
        
    def test_determine_ideals(self):
        self.cal.run()
        
        self.assertEqual(self.actuals[0], self.cal.ideals[0])
        self.assertEqual(self.actuals[1], self.cal.ideals[1])
            
    def test_from_coefs(self):
        raise SkipTest('not applicable')
    def test_from_coefs_ntwks(self):
        raise SkipTest('not applicable ')


class EightTermTest(unittest.TestCase, CalibrationTest):
    def setUp(self):
        self.n_ports = 2
        self.wg =WG
        wg= self.wg

        self.X = wg.random(n_ports =2, name = 'X')
        self.Y = wg.random(n_ports =2, name='Y')
        #Isolation terms
        self.If = wg.random(n_ports=1, name='If')
        self.Ir = wg.random(n_ports=1, name='Ir')
        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')

        ideals = [
            wg.short(nports=2, name='short'),
            wg.open(nports=2, name='open'),
            wg.match(nports=2, name='load'),
            wg.thru(name='thru'),
            ]
            
        measured = [self.measure(k) for k in ideals]
        
        self.cal = rf.EightTerm(
            ideals = ideals,
            measured = measured,
            isolation = measured[2],
            switch_terms = (self.gamma_f, self.gamma_r)
            )

    def terminate(self, ntwk):
        '''
        terminate a measured network with the switch terms
        '''
        return terminate(ntwk,self.gamma_f, self.gamma_r)
        
    def measure(self,ntwk):
        out =  self.terminate(self.X**ntwk**self.Y)
        out.name = ntwk.name
        out.s[:,1,0] += self.If.s[:,0,0]
        out.s[:,0,1] += self.Ir.s[:,0,0]
        return out

    def test_unterminating(self):
        a = self.wg.random(n_ports=self.n_ports)
        #unermintated measurment
        ut =  self.X**a**self.Y
        #terminated measurement
        m = self.measure(a)
        #Remove leakage
        m.s[:,1,0] -= self.If.s[:,0,0]
        m.s[:,0,1] -= self.Ir.s[:,0,0]
        self.assertEqual(self.cal.unterminate(m), ut)

    def test_forward_directivity_accuracy(self):
        self.assertEqual(
            self.X.s11,
            self.cal.coefs_ntwks['forward directivity'])
    
    def test_forward_source_match_accuracy(self):
        self.assertEqual(
            self.X.s22 , 
            self.cal.coefs_ntwks['forward source match'] )       
    
    def test_forward_reflection_tracking_accuracy(self):
        self.assertEqual(
            self.X.s21 * self.X.s12 , 
            self.cal.coefs_ntwks['forward reflection tracking'])
    
    def test_reverse_source_match_accuracy(self):
        self.assertEqual(
            self.Y.s11 , 
            self.cal.coefs_ntwks['reverse source match']   )     
    
    def test_reverse_directivity_accuracy(self):
        self.assertEqual(
            self.Y.s22 , 
            self.cal.coefs_ntwks['reverse directivity']  )      
    
    def test_reverse_reflection_tracking_accuracy(self):
        self.assertEqual(
            self.Y.s21 * self.Y.s12 , 
            self.cal.coefs_ntwks['reverse reflection tracking'])
    
    def test_k_accuracy(self):
        self.assertEqual(
            self.X.s21/self.Y.s12 , 
            self.cal.coefs_ntwks['k']  )   

    def test_forward_isolation_accuracy(self):
        self.assertEqual(
            self.If.s11 , 
            self.cal.coefs_ntwks['forward isolation']  )      

    def test_reverse_isolation_accuracy(self):
        self.assertEqual(
            self.Ir.s11 , 
            self.cal.coefs_ntwks['reverse isolation']  )      

    def test_verify_12term(self):
        self.assertTrue(self.cal.verify_12term_ntwk.s_mag.max() < 1e-3)


class TRLTest(EightTermTest):
    def setUp(self):
        self.n_ports = 2
        self.wg = WG
        wg= self.wg
        
        
        self.X = wg.random(n_ports =2, name = 'X')
        self.Y = wg.random(n_ports =2, name='Y')
        self.If = wg.random(n_ports=1, name='If')
        self.Ir = wg.random(n_ports=1, name='Ir')
        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')
        # make error networks have s21,s12 >> s11,s22 so that TRL
        # can guess at line length
        self.X.s[:,0,0] *=1e-1
        self.Y.s[:,0,0] *=1e-1
        self.X.s[:,1,1] *=1e-1 
        self.Y.s[:,1,1] *=1e-1 
        
        actuals = [
            wg.thru( name='thru'),
            rf.two_port_reflect(wg.load(-.9-.1j),wg.load(-.9-.1j)),
            wg.attenuator(-3,True, 45,'deg')
            #wg.line(45,'deg',name='line'),
            ]
        self.actuals=actuals
        ideals = [
            wg.thru( name='thru'),
            wg.short(nports=2, name='short'),
            wg.line(90,'deg',name='line'),
            ]
            
        measured = [self.measure(k) for k in actuals]
        
        self.cal = rf.TRL(
            ideals = ideals,
            measured = measured,
            isolation = measured[1],
            switch_terms = (self.gamma_f, self.gamma_r)
            )

    def test_found_line(self):
        self.cal.run()
        self.assertTrue(self.cal.ideals[2]==self.actuals[2])
        
    def test_found_reflect(self):
        self.cal.run()
        self.assertTrue(self.cal.ideals[1]==self.actuals[1])


class TRLWithNoIdealsTest(EightTermTest):
    def setUp(self):
        self.n_ports = 2
        self.wg = WG
        wg= self.wg

        self.X = wg.random(n_ports =2, name = 'X')
        self.Y = wg.random(n_ports =2, name='Y')
        self.If = wg.random(n_ports=1, name='If')
        self.Ir = wg.random(n_ports=1, name='Ir')
        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')
        # make error networks have s21,s12 >> s11,s22 so that TRL
        # can guess at line length
        self.X.s[:,0,0] *=1e-1
        self.Y.s[:,0,0] *=1e-1
        self.X.s[:,1,1] *=1e-1 
        self.Y.s[:,1,1] *=1e-1 
        
        ideals =  None
        
        r = wg.delay_short(20,'deg')

        self.actuals=[wg.thru( name='thru'),
                      rf.two_port_reflect(r,r),\
                      wg.attenuator(-3,True, 45,'deg')]

        measured = [self.measure(k) for k in self.actuals]
        
        self.cal = rf.TRL(
            ideals = None,
            measured = measured,
            isolation = measured[1],
            switch_terms = (self.gamma_f, self.gamma_r)
            )

    def test_found_line(self):
        self.cal.run()
        self.assertTrue(self.cal.ideals[2]==self.actuals[2])
        
    def test_found_reflect(self):
        self.cal.run()
        self.assertTrue(self.cal.ideals[1]==self.actuals[1])


class TRLMultiline(EightTermTest):
    def setUp(self):
        self.n_ports = 2
        self.wg = WG
        wg= self.wg

        self.X = wg.random(n_ports =2, name = 'X')
        self.Y = wg.random(n_ports =2, name='Y')
        self.If = wg.random(n_ports=1, name='If')
        self.Ir = wg.random(n_ports=1, name='Ir')
        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')
        # make error networks have s21,s12 >> s11,s22 so that TRL
        # can guess at line length
        #self.X.s[:,0,0] *=1e-1
        #self.Y.s[:,0,0] *=1e-1
        #self.X.s[:,1,1] *=1e-1 
        #self.Y.s[:,1,1] *=1e-1 
        
        ideals =  None
        
        actuals = [
            wg.thru( name='thru'),
            wg.short(nports=2, name='short'),
            wg.open(nports=2, name='open'),
            wg.attenuator(-3,True, 45,'deg'),
            wg.attenuator(-6,True, 90,'deg'),
            wg.attenuator(-8,True, 145,'deg'),
            ]
        self.actuals=actuals

        measured = [self.measure(k) for k in actuals]
        
        self.cal = rf.TRL(
            ideals = [None, -1,1,None,None,None],
            measured = measured,
            isolation = measured[1],
            switch_terms = (self.gamma_f, self.gamma_r),
            n_reflects=2,
            )

    def test_found_line(self):
        self.cal.run()
        for k in range(2,5):
            self.assertTrue(self.cal.ideals[k]==self.actuals[k])         


class NISTMultilineTRLTest(EightTermTest):
    def setUp(self):
        self.n_ports = 2
        self.wg = WG
        wg= self.wg

        self.X = wg.random(n_ports =2, name = 'X')
        self.Y = wg.random(n_ports =2, name = 'Y')
        self.If = wg.random(n_ports=1, name='If')
        self.Ir = wg.random(n_ports=1, name='Ir')

        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')

        actuals = [
            wg.thru(),
            rf.two_port_reflect(wg.load(-.98-.1j),wg.load(-.98-.1j)),
            rf.two_port_reflect(wg.load(.99+0.05j),wg.load(.99+0.05j)),
            wg.line(100,'um'),
            wg.line(200,'um'),
            wg.line(900,'um'),
            ]

        self.actuals=actuals

        measured = [self.measure(k) for k in actuals]

        self.cal = NISTMultilineTRL(
            measured = measured,
            isolation = measured[1],
            Grefls = [-1, 1],
            l = [0, 100e-6, 200e-6, 900e-6],
            er_est = 1,
            switch_terms = (self.gamma_f, self.gamma_r),
            gamma_root_choice = 'real'
            )

    def test_gamma(self):
        self.assertTrue(max(npy.abs(self.wg.gamma-self.cal.gamma)) < 1e-3)


class NISTMultilineTRLTest2(unittest.TestCase):
    """ Test characteristic impedance change and reference plane shift.
    Due to the transformations solved error boxes are not equal to the initial
    error boxes so CalibrationTestCase can't be used."""
    def setUp(self):
        global NPTS
        self.n_ports = 2
        self.wg = WG
        wg = self.wg

        r = npy.random.uniform(10,100,NPTS)
        l = 1e-9*npy.random.uniform(100,200,NPTS)
        g = npy.zeros(NPTS)
        c = 1e-12*npy.random.uniform(100,200,NPTS)

        rlgc = DistributedCircuit(frequency=wg.frequency, z0=None, R=r, L=l, G=g, C=c)
        self.rlgc = rlgc

        self.X = wg.random(n_ports =2, name = 'X')
        self.Y = wg.random(n_ports =2, name = 'Y')
        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')

        actuals = [
            rlgc.thru(),
            rlgc.short(nports=2),
            rlgc.line(10,'um'),
            rlgc.line(100,'um'),
            rlgc.line(500,'um'),
            ]

        self.actuals=actuals

        measured = [self.measure(k) for k in actuals]

        self.measured = measured

        self.cal = NISTMultilineTRL(
            measured = measured,
            Grefls = [-1],
            l = [0, 10e-6, 100e-6, 500e-6],
            switch_terms = (self.gamma_f, self.gamma_r),
            ref_plane=50e-6,
            c0=c,
            z0_ref=50,
            gamma_root_choice = 'real'
            )

    def terminate(self, ntwk):
        '''
        terminate a measured network with the switch terms
        '''
        return terminate(ntwk,self.gamma_f, self.gamma_r)

    def measure(self,ntwk):
        out =  self.terminate(self.X**ntwk**self.Y)
        out.name = ntwk.name
        return out

    def test_gamma(self):
        self.assertTrue(max(npy.abs(self.rlgc.gamma-self.cal.gamma)) < 1e-3)

    def test_z0(self):
        self.assertTrue(max(npy.abs(self.rlgc.z0-self.cal.z0)) < 1e-3)

    def test_shift(self):
        npy.testing.assert_allclose(self.cal.apply_cal(self.measured[3]).s, 
                                    self.wg.thru().s, atol=1e-11)
        
    def test_shift2(self):
        feed = self.rlgc.line(50,'um')
        dut = self.wg.random(n_ports=2)
        #Thrus convert the port impedances to 50 ohm
        dut_feed = self.wg.thru()**feed**dut**feed**self.wg.thru()
        dut_meas = self.measure(dut_feed)
        npy.testing.assert_allclose(self.cal.apply_cal(dut_meas).s, 
                                    dut.s, atol=1e-11)

class TREightTermTest(unittest.TestCase, CalibrationTest):
    def setUp(self):
        raise SkipTest()
        self.n_ports = 2
        self.wg = WG
        wg= self.wg
        
        
        self.X = wg.random(n_ports =2, name = 'X')
        self.Y = wg.random(n_ports =2, name='Y')
        
        
        ideals = [
            wg.short(nports=2, name='short'),
            wg.open(nports=2, name='open'),
            wg.match(nports=2, name='load'),
            wg.thru(name='thru'),
            ]
            
        measured = [self.measure_std(k) for k in ideals]
        
        cal1 = rf.TwoPortOnePath(
            ideals = ideals,
            measured = measured
            )
        switch_terms = (cal1.coefs_ntwks['forward switch term'],
                        cal1.coefs_ntwks['reverse switch term'])
        
        
        measured = [self.measure(k) for k in ideals]
        self.cal = rf.EightTerm(
            ideals = ideals,
            measured = measured,
            switch_terms = switch_terms, 
            )
        raise ValueError()
        
    def measure_std(self,ntwk):
        r= self.wg.random(2)
        m = ntwk.copy()
        mf = self.X**ntwk**self.Y
        
        m.s[:,1,0] = mf.s[:,1,0]
        m.s[:,0,0] = mf.s[:,0,0]
        m.s[:,1,1] = r.s[:,1,1]
        m.s[:,0,1] = r.s[:,0,1]
        return m    
    def measure(self,ntwk):
        
        m = ntwk.copy()
        mf = self.X**ntwk**self.Y
        mr = self.X**ntwk.flipped()**self.Y
        
        m.s[:,1,0] = mf.s[:,1,0]
        m.s[:,0,0] = mf.s[:,0,0]
        m.s[:,1,1] = mr.s[:,0,0]
        m.s[:,0,1] = mr.s[:,1,0]
        return m


class TwelveTermTest(unittest.TestCase, CalibrationTest):
    '''
    This test verifys the accuracy of the SOLT calibration. Generating 
    measured networks requires different error networks for forward and 
    reverse excitation states, these are described as follows
    
    forward excitation
        used for S21 and S11
        Mf = Xf ** S ** Yf  
    
    reverse excitation
        used for S12 and S22
        Mr = Xr ** S ** Yr
    
    
    '''
    def setUp(self):
        self.n_ports = 2
        self.wg = WG
        wg  = self.wg
        self.Xf = wg.random(n_ports =2, name = 'Xf')
        self.Xr = wg.random(n_ports =2, name = 'Xr')
        self.Yf = wg.random(n_ports =2, name='Yf')
        self.Yr = wg.random(n_ports =2, name='Yr')
        self.If = wg.random(n_ports =1, name='If')
        self.Ir = wg.random(n_ports =1, name='Ir')
       
        ideals = [
            wg.short(nports=2, name='short'),
            wg.open(nports=2, name='open'),
            wg.match(nports=2, name='load'),
            wg.random(2,name='rand1'),
            wg.random(2,name='rand2'),
            ]

        measured = [ self.measure(k) for k in ideals]
        
        self.cal = rf.TwelveTerm(
            ideals = ideals,
            measured = measured,
            n_thrus=2,
            isolation=measured[2]
            )
    
    def measure(self,ntwk):
        m = ntwk.copy()
        mf = self.Xf**ntwk**self.Yf
        mr = self.Xr**ntwk**self.Yr
        m.s[:,1,0] = mf.s[:,1,0]
        m.s[:,0,0] = mf.s[:,0,0]
        m.s[:,0,1] = mr.s[:,0,1]
        m.s[:,1,1] = mr.s[:,1,1]

        #Leakage
        m.s[:,1,0] += self.If.s[:,0,0]
        m.s[:,0,1] += self.Ir.s[:,0,0]
        return m

    def test_forward_directivity_accuracy(self):
        self.assertEqual(
            self.Xf.s11,
            self.cal.coefs_ntwks['forward directivity'])
    
    def test_forward_source_match_accuracy(self):
        self.assertEqual(
            self.Xf.s22 , 
            self.cal.coefs_ntwks['forward source match'] )       
    
    def test_forward_load_match_accuracy(self):
        self.assertEqual(
            self.Yf.s11 , 
            self.cal.coefs_ntwks['forward load match'])
    
    def test_forward_reflection_tracking_accuracy(self):
        self.assertEqual(
            self.Xf.s21 * self.Xf.s12 , 
            self.cal.coefs_ntwks['forward reflection tracking'])
    
    def test_forward_transmission_tracking_accuracy(self):
        self.assertEqual(
            self.Xf.s21*self.Yf.s21 , 
            self.cal.coefs_ntwks['forward transmission tracking'])
    
    def test_reverse_source_match_accuracy(self):
        self.assertEqual(
            self.Yr.s11 , 
            self.cal.coefs_ntwks['reverse source match']   )     
    
    def test_reverse_directivity_accuracy(self):
        self.assertEqual(
            self.Yr.s22 , 
            self.cal.coefs_ntwks['reverse directivity']  )      
    
    def test_reverse_load_match_accuracy(self):
        self.assertEqual(
            self.Xr.s22 , 
            self.cal.coefs_ntwks['reverse load match'])
    
    def test_reverse_reflection_tracking_accuracy(self):
        self.assertEqual(
            self.Yr.s21 * self.Yr.s12 , 
            self.cal.coefs_ntwks['reverse reflection tracking'])
    
    def test_reverse_transmission_tracking_accuracy(self):
        self.assertEqual(
            self.Yr.s12*self.Xr.s12 , 
            self.cal.coefs_ntwks['reverse transmission tracking'])
            
    def test_forward_isolation_accuracy(self):
        self.assertEqual(
            self.If.s11,
            self.cal.coefs_ntwks['forward isolation'])

    def test_reverse_isolation_accuracy(self):
        self.assertEqual(
            self.Ir.s11,
            self.cal.coefs_ntwks['reverse isolation'])
    
    @nottest
    def test_convert_12term_2_8term(self):
        converted = rf.convert_8term_2_12term(
                    rf.convert_12term_2_8term(self.cal.coefs))
        
        
        for k in converted:
            print(('{}-{}'.format(k,abs(self.cal.coefs[k] - converted[k]))))
        for k in converted:
            self.assertTrue(abs(self.cal.coefs[k] - converted[k])<1e-9)
        
    @nottest
    def test_convert_12term_2_8term_correction_accuracy(self):
        converted = rf.convert_8term_2_12term(
                    rf.convert_12term_2_8term(self.cal.coefs))
        
        self.cal._coefs = converted
        a = self.wg.random(n_ports=2)
        m = self.measure(a)
        c = self.cal.apply_cal(m)
               
        self.assertEqual(a,c)
    
    @nottest
    def test_verify_12term(self):
        
        self.assertTrue(self.cal.verify_12term_ntwk.s_mag.max() < 1e-3)


class TwelveTermSloppyInitTest(TwelveTermTest):
    '''
    Test the TwelveTerm.__init__'s ability to 
    1) determine the number of thrus (n_thrus) hueristically
    2) put the standards in correct order if they use sloppy_input
    
    It must be a entirely seperate test because we want to ensure it 
    creates an accurate calibration.
    '''
    @suppress_warning_decorator("dictionary passed, sloppy_input")
    @suppress_warning_decorator("n_thrus is None")
    def setUp(self):
        self.n_ports = 2
        self.wg = WG
        wg  = self.wg
        self.Xf = wg.random(n_ports =2, name = 'Xf')
        self.Xr = wg.random(n_ports =2, name = 'Xr')
        self.Yf = wg.random(n_ports =2, name='Yf')
        self.Yr = wg.random(n_ports =2, name='Yr')
        #No leakage as it can interfere with thru detection, which is done
        #based on S21 and S12
        self.If = wg.match(n_ports =1, name='If')
        self.Ir = wg.match(n_ports =1, name='Ir')
       
        ideals = [
            wg.short(nports=2, name='short'),
            wg.open(nports=2, name='open'),
            wg.match(nports=2, name='load'),
            wg.attenuator(-20,name='atten'),
            wg.line(45,'deg',name = 'line') ,          
            ]
        
    
        measured = [ self.measure(k) for k in ideals]

        self.cal= TwelveTerm(
            ideals = NetworkSet(ideals).to_dict(), 
            measured = NetworkSet(measured).to_dict(),
            n_thrus=None,
            )
     
    def measure(self,ntwk):
        m = ntwk.copy()
        mf = self.Xf**ntwk**self.Yf
        mr = self.Xr**ntwk**self.Yr
        m.s[:,1,0] = mf.s[:,1,0]
        m.s[:,0,0] = mf.s[:,0,0]
        m.s[:,0,1] = mr.s[:,0,1]
        m.s[:,1,1] = mr.s[:,1,1]
        return m    


class SOLTTest(TwelveTermTest):
    def setUp(self):
        self.n_ports = 2
        self.wg = WG
        wg  = self.wg
        self.Xf = wg.random(n_ports =2, name = 'Xf')
        self.Xr = wg.random(n_ports =2, name = 'Xr')
        self.Yf = wg.random(n_ports =2, name='Yf')
        self.Yr = wg.random(n_ports =2, name='Yr')
        self.If = wg.random(n_ports =1, name='If')
        self.Ir = wg.random(n_ports =1, name='Ir')

        ideals = [
            wg.short(nports=2, name='short'),
            wg.open(nports=2, name='open'),
            wg.match(nports=2, name='load'),
            None,            
            ]
        actuals = [
            wg.short(nports=2, name='short'),
            wg.open(nports=2, name='open'),
            wg.match(nports=2, name='load'),
            wg.thru(),            
            ]
    
        measured = [ self.measure(k) for k in actuals]
        
        self.cal = SOLT(
            ideals = ideals,
            measured = measured,
            n_thrus=1,
            isolation=measured[2]
            )


class TwoPortOnePathTest(TwelveTermTest):
    @suppress_warning_decorator("divide by zero encountered in log10")
    @suppress_warning_decorator("n_thrus is None")
    def setUp(self):
        self.n_ports = 2
        self.wg =WG
        wg  = self.wg
        self.Xf = wg.random(n_ports =2, name = 'Xf')
        self.Yf = wg.random(n_ports =2, name='Yf')

        #No leakage
        self.If = wg.match(n_ports =1, name='If')
        self.Ir = wg.match(n_ports =1, name='Ir')
        
        
        ideals = [
            wg.short(nports=2, name='short'),
            wg.open(nports=2, name='open'),
            wg.match(nports=2, name='load'),
            wg.random(2,name='rand1'),
            wg.random(2,name='rand2'),
            ]
        
    
        measured = [ self.measure(k) for k in ideals]
        self.cal = TwoPortOnePath(
            ideals = ideals,
            measured = measured,
            source_port=1,
            #n_thrus=2,
            )

    def measure(self,ntwk):
        r= self.wg.random(2)
        m = ntwk.copy()
        mf = self.Xf**ntwk**self.Yf
        
        m.s[:,1,0] = mf.s[:,1,0]
        m.s[:,0,0] = mf.s[:,0,0]
        m.s[:,1,1] = r.s[:,1,1]
        m.s[:,0,1] = r.s[:,0,1]
        return m

    def test_accuracy_of_dut_correction(self):
        a = self.wg.random(n_ports=self.n_ports, name = 'actual')
        f = self.measure(a)
        r = self.measure(a.flipped())
        c = self.cal.apply_cal((f,r))
        c.name = 'corrected'   
        self.assertEqual(c,a)
        
    def test_embed_then_apply_cal(self):
        a = self.wg.random(n_ports=self.n_ports)
        f = self.cal.embed(a)
        r = self.cal.embed(a.flipped())
        self.assertEqual(self.cal.apply_cal((f,r)),a)
        
    def test_embed_equal_measure(self):
        # measurment procedure is different so this test doesnt apply
        raise SkipTest()

    @suppress_warning_decorator("n_thrus is None")
    def test_from_coefs(self):
        cal_from_coefs = self.cal.from_coefs(self.cal.frequency, self.cal.coefs)
        ntwk = self.wg.random(n_ports=self.n_ports)
    
    @suppress_warning_decorator("n_thrus is None")
    def test_from_coefs_ntwks(self):
        cal_from_coefs = self.cal.from_coefs_ntwks(self.cal.coefs_ntwks)

    def test_reverse_source_match_accuracy(self):
        raise SkipTest()   
    
    def test_reverse_directivity_accuracy(self):
        raise SkipTest()      
    
    def test_reverse_load_match_accuracy(self):
        raise SkipTest()  
    
    def test_reverse_reflection_tracking_accuracy(self):
        raise SkipTest()  
    
    def test_reverse_transmission_tracking_accuracy(self):
        raise SkipTest()  


class UnknownThruTest(EightTermTest):
    def setUp(self):
        
        self.n_ports = 2
        self.wg = WG
        wg= self.wg 
        self.X = wg.random(n_ports =2, name = 'X')
        self.Y = wg.random(n_ports =2, name='Y')
        #No leakage
        self.If = wg.match(n_ports=1, name='If')
        self.Ir = wg.match(n_ports=1, name='Ir')
        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')
        
        
        actuals = [
            wg.short(nports=2, name='short'),
            wg.open(nports=2, name='open'),
            wg.match(nports=2, name='match'),
            wg.impedance_mismatch(50,45)**wg.line(20,'deg',name='line')**wg.impedance_mismatch(45,50)

            ]
        
        ideals = [
            wg.short(nports=2, name='short'),
            wg.open(nports=2, name='open'),
            wg.match(nports=2, name='match'),
            wg.thru(name='thru'),
            ]
            
        measured = [self.measure(k) for k in actuals]
        
        self.cal = rf.UnknownThru(
            ideals = ideals,
            measured = measured,
            switch_terms = [self.gamma_f, self.gamma_r]
            )


class MRCTest(EightTermTest):
    def setUp(self):
        
        self.n_ports = 2
        #Exact only with a lossless waveguide
        self.wg = WG_lossless
        wg= self.wg 
        self.X = wg.random(n_ports =2, name = 'X')
        self.Y = wg.random(n_ports =2, name='Y')
        #No leakage
        self.If = wg.match(n_ports=1, name='If')
        self.Ir = wg.match(n_ports=1, name='Ir')
        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')
        
        
        def delay_shorts(d1,d2):
            ds1 = wg.delay_short(d1,'deg')
            ds2 = wg.delay_short(d2,'deg')
            return rf.two_port_reflect(ds1,ds2)
        
        actuals = [
            wg.short(nports=2, name='short'),
            delay_shorts(65,130),
            delay_shorts(120,75),
            wg.load(.2+.2j,nports=2, name='match'),
            wg.impedance_mismatch(50,45)**wg.line(20,'deg',name='line')**wg.impedance_mismatch(45,50)

            ]
        
        ideals = [
            wg.short(nports=2, name='short'),
            delay_shorts(45,90),
            delay_shorts(90,45),
            wg.load(.2+.2j,nports=2, name='match'),
            wg.thru(name='thru'),
            ]
            
        measured = [self.measure(k) for k in actuals]
        
        self.cal = rf.MRC(
            ideals = ideals,
            measured = measured,
            switch_terms = [self.gamma_f, self.gamma_r]
            )


class TwelveTermToEightTermTest(unittest.TestCase, CalibrationTest):
    '''
    This test verifies the accuracy of the SOLT calibration, when used 
    on an error-box (8-term) model.
    
    
    '''
    @suppress_warning_decorator("n_thrus is None")
    def setUp(self):
        self.n_ports = 2
        wg= rf.wr10
        wg.frequency = rf.F.from_f([100])
        self.wg = wg
        self.X = wg.random(n_ports =2, name = 'X')
        self.Y = wg.random(n_ports =2, name='Y')
        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')
        #Isolation terms
        self.If = wg.random(n_ports=1, name='If')
        self.Ir = wg.random(n_ports=1, name='Ir')

        ideals = [
            wg.short(nports=2, name='short'),
            wg.open(nports=2, name='open'),
            wg.match(nports=2, name='load'),
            wg.thru(name='thru'),
            ]

        measured = [ self.measure(k) for k in ideals]
        with warnings.catch_warnings(record=False):
            self.cal = rf.TwelveTerm(
            ideals = ideals,
            measured = measured,
            isolation=measured[2]
            )


        coefs = rf.calibration.convert_12term_2_8term(self.cal.coefs, redundant_k=1)
        coefs = NetworkSet.from_s_dict(coefs,
                                    frequency=self.cal.frequency).to_dict()
        self.coefs= coefs

    def terminate(self, ntwk):
        '''
        terminate a measured network with the switch terms
        '''
        return terminate(ntwk,self.gamma_f, self.gamma_r)

    def measure(self,ntwk):
        out =  self.terminate(self.X**ntwk**self.Y)
        out.name = ntwk.name
        out.s[:,1,0] += self.If.s[:,0,0]
        out.s[:,0,1] += self.Ir.s[:,0,0]
        return out

    def test_forward_isolation(self):
        self.assertEqual(self.coefs['forward isolation'], self.If.s11)

    def test_reverse_isolation(self):
        self.assertEqual(self.coefs['reverse isolation'], self.Ir.s11)

    def test_forward_switch_term(self):
        self.assertEqual(self.coefs['forward switch term'], self.gamma_f)

    def test_reverse_switch_term(self):
        self.assertEqual(self.coefs['reverse switch term'], self.gamma_r)

    def test_forward_directivity_accuracy(self):
        self.assertEqual(
            self.X.s11,
            self.coefs['forward directivity'])

    def test_forward_source_match_accuracy(self):
        self.assertEqual(
            self.X.s22 ,
            self.coefs['forward source match'] )

    def test_forward_reflection_tracking_accuracy(self):
        self.assertEqual(
            self.X.s21 * self.X.s12 ,
            self.coefs['forward reflection tracking'])

    def test_reverse_source_match_accuracy(self):
        self.assertEqual(
            self.Y.s11 ,
            self.coefs['reverse source match']   )

    def test_reverse_directivity_accuracy(self):
        self.assertEqual(
            self.Y.s22 ,
            self.coefs['reverse directivity']  )

    def test_reverse_reflection_tracking_accuracy(self):
        self.assertEqual(
            self.Y.s21 * self.Y.s12 ,
            self.coefs['reverse reflection tracking'])

    def test_k_accuracy(self):
        self.assertEqual(
            self.X.s21/self.Y.s12 ,
            self.coefs['k']  )

    def test_verify_12term(self):
        self.assertTrue(self.cal.verify_12term_ntwk.s_mag.max() < 1e-3)


class SixteenTermTest(unittest.TestCase, CalibrationTest):
    def setUp(self):
        self.n_ports = 2
        self.wg = WG
        wg = self.wg

        #Port 0: VNA port 0
        #Port 1: DUT port 0
        #Port 2: DUT port 1
        #Port 3: VNA port 1
        self.Z = wg.random(n_ports = 4, name = 'Z')
        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')

        o = wg.open(nports=1, name='open')
        s = wg.short(nports=1, name='short')
        m = wg.match(nports=1, name='load')
        om = rf.two_port_reflect(o, m)
        mo = rf.two_port_reflect(m, o)
        oo = rf.two_port_reflect(o, o)
        ss = rf.two_port_reflect(s, s)
        thru = wg.thru(name='thru')

        ideals = [
            thru,
            om,
            mo,
            oo,
            ss
            ]

        measured = [self.measure(k) for k in ideals]

        self.cal = rf.SixteenTerm(
            measured = measured,
            ideals = ideals,
            switch_terms=(self.gamma_f, self.gamma_r)
            )

    def terminate(self, ntwk):
        '''
        terminate a measured network with the switch terms
        '''
        return terminate(ntwk,self.gamma_f, self.gamma_r)

    def measure(self,ntwk):
        out = self.terminate(rf.connect(self.Z, 1, ntwk, 0, num=2))
        out.name = ntwk.name
        return out

    def test_forward_directivity_accuracy(self):
        self.assertEqual(
            self.Z.s11,
            self.cal.coefs_ntwks['forward directivity'])

    def test_forward_source_match_accuracy(self):
        self.assertEqual(
            self.Z.s22 ,
            self.cal.coefs_ntwks['forward source match'] )

    def test_forward_reflection_tracking_accuracy(self):
        self.assertEqual(
            self.Z.s21 * self.Z.s12 ,
            self.cal.coefs_ntwks['forward reflection tracking'])

    def test_reverse_source_match_accuracy(self):
        self.assertEqual(
            self.Z.s33 ,
            self.cal.coefs_ntwks['reverse source match']   )

    def test_reverse_directivity_accuracy(self):
        self.assertEqual(
            self.Z.s44 ,
            self.cal.coefs_ntwks['reverse directivity']  )

    def test_reverse_reflection_tracking_accuracy(self):
        self.assertEqual(
            self.Z.s34 * self.Z.s43 ,
            self.cal.coefs_ntwks['reverse reflection tracking'])

    def test_forward_isolation(self):
        self.assertEqual(
            self.Z.s41 ,
            self.cal.coefs_ntwks['forward isolation'])

    def test_reverse_isolation(self):
        self.assertEqual(
            self.Z.s14 ,
            self.cal.coefs_ntwks['reverse isolation'])

    def test_reverse_port_isolation(self):
        self.assertEqual(
            self.Z.s23 ,
            self.cal.coefs_ntwks['reverse port isolation'])

    def test_forward_port_isolation(self):
        self.assertEqual(
            self.Z.s32 ,
            self.cal.coefs_ntwks['forward port isolation'])

    def test_k(self):
        self.assertEqual(
            self.Z.s21/self.Z.s34 ,
            self.cal.coefs_ntwks['k'])

    def test_forward_port1_isolation(self):
        self.assertEqual(
            self.Z.s31/self.Z.s34 ,
            self.cal.coefs_ntwks['forward port 1 isolation'])

    def test_reverse_port1_isolation(self):
        self.assertEqual(
            self.Z.s13*self.Z.s34 ,
            self.cal.coefs_ntwks['reverse port 1 isolation'])

    def test_forward_port2_isolation(self):
        self.assertEqual(
            self.Z.s42*self.Z.s34 ,
            self.cal.coefs_ntwks['forward port 2 isolation'])

    def test_reverse_port2_isolation(self):
        self.assertEqual(
            self.Z.s24/self.Z.s34 ,
            self.cal.coefs_ntwks['reverse port 2 isolation'])

class SixteenTermCoefficientsTest(unittest.TestCase):
    """Test that 16-term non-isolation coefficients are defined the same way as
    8-term coefficients."""
    def setUp(self):
        self.n_ports = 2
        self.wg = WG
        wg = self.wg

        #Port 0: VNA port 0
        #Port 1: DUT port 0
        #Port 2: DUT port 1
        #Port 3: VNA port 1
        self.Z = wg.random(n_ports = 4, name = 'Z')
        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')

        o = wg.open(nports=1, name='open')
        s = wg.short(nports=1, name='short')
        m = wg.match(nports=1, name='load')
        om = rf.two_port_reflect(o, m)
        mo = rf.two_port_reflect(m, o)
        oo = rf.two_port_reflect(o, o)
        ss = rf.two_port_reflect(s, s)
        thru = wg.thru(name='thru')

        ideals = [
            thru,
            om,
            mo,
            oo,
            ss
            ]

        measured = [self.measure(k) for k in ideals]

        self.cal16 = rf.SixteenTerm(
            measured = measured,
            ideals = ideals,
            switch_terms=(self.gamma_f, self.gamma_r)
            )

        r = wg.load(.95+.1j,nports=1)
        m = wg.match(nports=1)
        mm = rf.two_port_reflect(m, m)
        rm = rf.two_port_reflect(r, m)
        mr = rf.two_port_reflect(m, r)
        rr = rf.two_port_reflect(r, r)

        ideals = [
            thru,
            mm,
            rr,
            rm,
            mr
            ]

        measured = [self.measure(k) for k in ideals]

        self.cal_lmr16 = rf.LMR16(
            measured = measured,
            ideals = [thru],
            ideal_is_reflect = False,
             #Automatic sign detection doesn't work if the
             #error terms aren't symmetric enough
            sign = 1,
            switch_terms=(self.gamma_f, self.gamma_r)
            )

        #Same error network, but without leakage terms

        #Primary leakage
        self.Z.s[:,3,0] = 0 # forward isolation
        self.Z.s[:,0,3] = 0 # reverse isolation
        self.Z.s[:,2,1] = 0 # forward port isolation
        self.Z.s[:,1,2] = 0 # reverse port isolation

        #Cross leakage        
        self.Z.s[:,3,1] = 0 # forward port 2 isolation
        self.Z.s[:,1,3] = 0 # reverse port 2 isolation
        self.Z.s[:,2,0] = 0 # forward port 1 isolation
        self.Z.s[:,0,2] = 0 # reverse port 1 isolation

        measured = [self.measure(k) for k in ideals]

        self.cal8 = rf.EightTerm(
            measured = measured,
            ideals = ideals,
            switch_terms=(self.gamma_f, self.gamma_r),
            )

    def terminate(self, ntwk):
        '''
        terminate a measured network with the switch terms
        '''
        return terminate(ntwk,self.gamma_f, self.gamma_r)

    def measure(self,ntwk):
        out = self.terminate(rf.connect(self.Z, 1, ntwk, 0, num=2))
        out.name = ntwk.name
        return out

    def test_coefficients(self):
        for k in self.cal8.coefs.keys():
            if k in self.cal16.coefs.keys():
                if 'isolation' in k:
                    continue
                self.assertTrue(all(npy.abs(self.cal8.coefs[k] - self.cal16.coefs[k]) < 1e-10))
                self.assertTrue(all(npy.abs(self.cal8.coefs[k] - self.cal_lmr16.coefs[k]) < 1e-10))


class LMR16Test(SixteenTermTest):
    def setUp(self):
        self.n_ports = 2
        self.wg = WG
        wg = self.wg

        #Port 0: VNA port 0
        #Port 1: DUT port 0
        #Port 2: DUT port 1
        #Port 3: VNA port 1
        self.Z = wg.random(n_ports = 4, name = 'Z')
        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')

        r = wg.short(nports=1, name='short')
        m = wg.match(nports=1, name='load')
        mm = rf.two_port_reflect(m, m)
        rm = rf.two_port_reflect(r, m)
        mr = rf.two_port_reflect(m, r)
        rr = rf.two_port_reflect(r, r)

        thru_length = uniform(0,10)
        thru = wg.line(thru_length,'deg',name='line')

        self.thru = thru

        ideals = [
            thru,
            mm,
            rr,
            rm,
            mr
            ]

        measured = [self.measure(k) for k in ideals]

        self.cal = rf.LMR16(
            measured = measured,
            ideals = [r],
            ideal_is_reflect = True,
             #Automatic sign detection doesn't work if the
             #error terms aren't symmetric enough
            sign = -1,
            switch_terms=(self.gamma_f, self.gamma_r)
            )

    def test_solved_through(self):
        self.assertEqual(
            self.thru,
            self.cal.solved_through)

    
if __name__ == "__main__":
    unittest.main()





from traits.api import *
from traitsui.api import *#View, Item, ButtonEditor, Group, HSplit,
from traitsui.menu import *

import numpy as npy
import os
import pylab as plb
import skrf as rf




class PlotTool(HasTraits):
    save_dir = Directory()
    clf = Button('Clear')
    newfig = Button('New')
    close_all = Button('Close All')
    tight_layout = Button('Tight Layout')
    draw_show = Button('Draw/Show')
    save_all_figs = Button('Save All')
    plot_format =  Enum('png',['svg','eps','jpg','png'],label='Format')
    title = String('')
    apply_title = Button('Title')
    def _clf_fired(self):
        plb.clf()
        plb.draw()
        plb.show()

    def _newfig_fired(self):
        plb.figure()
        plb.draw()
        plb.show()

    def _close_all_fired(self):
        plb.close('all')

    def _tight_layout_fired(self):
        plb.tight_layout()
        plb.draw()
        plb.show()

    def _draw_show_fired(self):
        plb.draw()
        plb.show()

    def _save_all_figs_fired(self):
        rf.save_all_figs(dir=self.save_dir, format = [self.plot_format])
        plb.ion()

    def _apply_title_fired(self):
        plb.title(self.title)
        plb.draw();plb.show()

    view = View(
        Group(
            HSplit(
                Item('newfig', label='Figure\t\t'),
                Item('clf', show_label=False),
                Item('close_all', show_label=False),
                Item('tight_layout', show_label=False),
                Item('draw_show', show_label=False),
                ),
            HSplit(
                Item('apply_title', label='Formatting\t'),
                Item('title', show_label=False),
                ),
            HSplit(
                Item('save_all_figs', label='Save\t\t'),
                Item('plot_format'),
                Item('save_dir', label='Directory'),

                )
            )
        )



import skrf as rf
import unittest
from numpy import real, imag, linspace, pi, tanh, array, conj
from numpy.random import rand
from numpy.testing import assert_equal, run_module_suite, assert_almost_equal


class TestBasicTransmissionLine(unittest.TestCase):
    """
    Test reflection coefficient-related functions
    """

    def setUp(self):
        # define example test case
        self.d = 1.5  # line length [m]
        self.z0 = 100  # line characteric impedance [Ohm]
        self.zin = 40 - 280j  # line input impedance [Ohm]
        
    def test_input_reflection_coefficient(self):
        """
        Test the input reflection coefficient value of the line.        
        """
        Gamma_in = rf.zl_2_Gamma0(self.z0, self.zin)
        
        expected_Gamma_in = (self.zin - self.z0)/(self.zin + self.z0)
        
        assert_equal(Gamma_in, expected_Gamma_in)

    def test_propagation_constant_from_reflection_coefficient(self):
        """
        Test the propagation constant value deduced from reflection coef
        """
        Gamma_in = rf.zl_2_Gamma0(self.z0, self.zin)
        Gamma_l = -1  # short
        
        gamma = rf.reflection_coefficient_2_propagation_constant(Gamma_in, 
                                                                 Gamma_l, 
                                                                 self.d)
        expected_gamma = 0.02971 + 1.272j
        
        assert_almost_equal(real(gamma), real(expected_gamma), decimal=4)
        assert_almost_equal(imag(gamma), imag(expected_gamma), decimal=4)
        

class ElectricalLengthTests(unittest.TestCase):
    """
    Test the functions related to electrical length conversions.
    """
    def setUp(self):
        self.d = 1.5  # m
        self.gamma0 = 0.2 + 5j
        self.gammas = 0.2 + 1j*linspace(1, 10, num=50)
        self.f0 = 50e6
        self.fs = linspace(1, 50, num=50)*1e6
        self.theta0 = self.gamma0 * self.d
        self.thetas = self.gammas * self.d

        
    def gamma_from_f(self, f0):
        """
        Dummy gamma(f) 

        Parameters
        ----------
        f0 : number of array-like
            frequency in Hz

        Returns
        -------
        gamma : number of array-like
            propagation constant

        """
        return (0.2 + 5j)*f0
    
    def test_electrical_length_from_length(self):
        """
        Test the conversions from physical distance to electrical lengths.
        """
        # test for gamma passed as scalar 
        theta_scalar = rf.electrical_length(self.gamma0, self.f0, self.d)
        theta_scalar_expected = self.gamma0 * self.d
        assert_almost_equal(theta_scalar, theta_scalar_expected)

        # test for gamma passed as array like 
        theta_array = rf.electrical_length(self.gammas, self.f0, self.d)
        theta_array_expected = self.gammas * self.d
        assert_almost_equal(theta_array, theta_array_expected)

        # test for gamma passed as function
        theta_function = rf.electrical_length(self.gamma_from_f, self.fs, self.d)
        theta_function_expected = self.gamma_from_f(self.fs) * self.d
        assert_almost_equal(theta_function, theta_function_expected)


    def test_length_from_electrical_distance(self):
        """
        Test the conversions from electrical length to physical distances. 
        """
        # test for gamma passed as scalar 
        d_scalar = rf.electrical_length_2_distance(self.theta0, self.gamma0, self.f0, deg=False)
        d_scalar_expected = real(self.theta0 / self.gamma0)
        assert_almost_equal(d_scalar, d_scalar_expected)

        # test for gamma passed as array-like
        d_array = rf.electrical_length_2_distance(self.theta0, self.gammas, self.f0, deg=False)
        d_array_expected = real(self.theta0 / self.gammas)
        assert_almost_equal(d_array, d_array_expected)

        # test for gamma passed as function 
        d_function = rf.electrical_length_2_distance(self.theta0, self.gamma_from_f, self.fs, deg=False)
        d_function_expected = real(self.theta0 / self.gamma_from_f(self.fs))
        assert_almost_equal(d_function, d_function_expected)
        # with theta passed as array
        d_function = rf.electrical_length_2_distance(self.thetas, self.gamma_from_f, self.fs, deg=False)
        d_function_expected = real(self.thetas / self.gamma_from_f(self.fs))
        assert_almost_equal(d_function, d_function_expected)
        
        # with theta passed in degrees
        d_scalar = rf.electrical_length_2_distance(self.theta0*180/pi, self.gamma0, self.f0, deg=True)
        d_scalar_expected = real(self.theta0 / self.gamma0)
        assert_almost_equal(d_scalar, d_scalar_expected)        
        
        
        
        
        
class TestVoltageCurrentPropagation(unittest.TestCase):
    def setUp(self):
        pass
    
    def test_d_zero(self):
        """
        Propagate voltage and current on a d=0 transmission line.
        Voltage and current are of course equal.
        """
        gamma = array([1j])
        d = 0
        z0 = 50
        v1 = 3
        i1 = 2
        theta = gamma * d
        
        v2, i2 = rf.voltage_current_propagation(v1, i1, z0, theta)
        
        assert_almost_equal(v2, v1)
        assert_almost_equal(i2, i1)
        
    def test_d_wavelength(self):
        """
        Propagate voltage and current on a d=lambda lossless transmission line.
        Voltage and current are equal.
        """
        gamma = array([1j])
        z0 = 50
        v1 = rand()
        i1 = rand()
        theta = 1j*2*pi
        
        v2, i2 = rf.voltage_current_propagation(v1, i1, z0, theta)
        assert_almost_equal(v2, v1)
        assert_almost_equal(i2, i1)

    def test_d_half_wavelength(self):
        """
        Propagate voltage and current on a d=lambda/2 lossless transmission line. 
        Voltage and current are inversed.
        """
        gamma = array([1j])
        z0 = 50
        v1 = rand()
        i1 = rand()
        theta = 1j*pi
        
        v2, i2 = rf.voltage_current_propagation(v1, i1, z0, theta)
        
        assert_almost_equal(v2, -v1)
        assert_almost_equal(i2, -i1)


if __name__ == "__main__":
    # Launch all tests
    run_module_suite()
    
